import re
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from supervisely import (
    Api,
    ObjClass,
    PointcloudAnnotation,
    PointcloudEpisodeAnnotation,
    PointcloudEpisodeFrame,
    PointcloudEpisodeObject,
    PointcloudFigure,
    ProjectMeta,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.bag.bag_helper import process_pc2_msg, process_vector3_msg
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    list_files_recursively,
    silent_remove,
)
from supervisely.project.project_settings import LabelingInterface


class BagConverter(PointcloudConverter):
    class Item(PointcloudConverter.Item):
        def __init__(
            self,
            item_path,
            ann_data: str = None,
            related_images: list = None,
            custom_data: dict = None,
        ):
            super().__init__(item_path, ann_data, related_images, custom_data)
            self._topic = None
            self._type = "point_cloud"

        @property
        def topic(self):
            return self._topic

        @topic.setter
        def topic(self, topic: str):
            self._topic = topic

    def __init__(
            self,
            input_data: str,
            labeling_interface: Optional[Union[LabelingInterface, str]],
            upload_as_links: bool,
            remote_files_map: Optional[Dict[str, str]] = None,
    ):
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self._total_msg_count = 0
        self._is_pcd_episode = False

    def __str__(self) -> str:
        return AvailablePointcloudConverters.BAG

    @property
    def key_file_ext(self) -> str:
        return ".bag"

    def validate_format(self) -> bool:
        import rosbag  # pylint: disable=import-error

        def _filter_fn(file_path):
            return get_file_ext(file_path).lower() == self.key_file_ext

        bag_files = list_files_recursively(self._input_data, filter_fn=_filter_fn)

        if len(bag_files) == 0:
            return False

        self._items = []
        for bag_file in bag_files:
            with rosbag.Bag(bag_file) as bag:
                cloud_msg_cnt = 0
                types, bag_msg_cnt = [], []
                topics_info = bag.get_type_and_topic_info()[1]

                for i in range(0, len(topics_info.values())):
                    types.append(list(topics_info.values())[i][0])
                    bag_msg_cnt.append(list(topics_info.values())[i][1])

                topics = zip(topics_info, types, bag_msg_cnt)
                pcd_topic = None
                ann_topic = None
                for topic, topic_type, msg_count in topics:
                    if topic_type == "sensor_msgs/PointCloud2":
                        if "SlyAnnotations" in topic:
                            logger.warn(
                                f"{topic} topic: only [geometry_msgs/Vector3Stamped] supported for Supervisely annotations"
                            )
                            # ann_topic = topic
                            continue
                        cloud_msg_cnt += msg_count
                        pcd_topic = topic
                    elif topic_type == "geometry_msgs/Vector3Stamped":
                        if "SlyAnnotations" in topic:
                            ann_topic = topic
                            continue

                if pcd_topic is not None:
                    if self._is_pcd_episode:
                        item = self.Item( # pylint: disable=unexpected-keyword-arg
                            item_path=bag_file, frame_number=cloud_msg_cnt
                        )  
                    else:
                        item = self.Item(item_path=bag_file)
                    item.topic = pcd_topic
                    item.ann_data = ann_topic
                    self._items.append(item)
                self._total_msg_count += cloud_msg_cnt

        return self.items_count > 0

    def convert(self, item: Item, meta: ProjectMeta):
        import rosbag  # pylint: disable=import-error

        bag_path = Path(item.path)
        topic = item.topic
        ann_topic = item.ann_data

        with rosbag.Bag(item.path) as bag:
            msg_count = bag.get_message_count(topic_filters=item.topic)
            if ann_topic is not None:
                ann_topics_info = bag.get_type_and_topic_info(topic_filters=[item.ann_data])[1]
                ann_topic_type = list(ann_topics_info.values())[0][0]
            progress, progress_cb = self.get_progress(
                msg_count, f"Convert {topic} topic from {bag_path.name} to pcd"
            )

            time_to_data = defaultdict(dict)

            for _, msg, rostime in bag.read_messages(topics=[item.topic]):
                process_pc2_msg(time_to_data, msg, rostime, bag_path, topic, meta, is_ann=False)

                progress_cb(1)
            if is_development():
                progress.close()

            # get annotations
            if ann_topic is not None:
                msg_count = bag.get_message_count(topic_filters=item.ann_data)
                progress, progress_cb = self.get_progress(
                    msg_count, f"Convert {item.ann_data} topic to JSON annotations"
                )

                time_to_vectors = defaultdict(list)
                for _, msg, rostime in bag.read_messages(topics=[item.ann_data]):
                    if ann_topic_type == "geometry_msgs/Vector3Stamped":
                        frame_id = msg.header.frame_id
                        if re.match(r"\d+\.\d+", frame_id):
                            time_to_vectors[frame_id].append(msg)
                    elif ann_topic_type == "sensor_msgs/PointCloud2":
                        logger.warn(
                            f"{item.ann_data} topic: only [geometry_msgs/Vector3Stamped] supported for Supervisely annotations"
                        )
                        # process_pc2_msg(time_to_data, msg, rostime, bag_path, ann_topic, meta, is_ann=True)
                        progress_cb(1)
                if len(time_to_vectors) > 0:
                    process_vector3_msg(time_to_data, time_to_vectors, bag_path, meta, ann_topic, progress_cb)

                if is_development():
                    progress.close()

        return time_to_data

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        self._upload_dataset(api, dataset_id, log_progress, is_episodes=self._is_pcd_episode)

    def _upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        log_progress=True,
        is_episodes=False,
    ):
        """
        Converts and uploads bag files to Supervisely dataset.
        Note: This method is used by both the BagConverter and the BagEpisodeConverter.
        """
        obj_cls = ObjClass("object", Cuboid3d)
        self._meta = ProjectMeta(obj_classes=[obj_cls])
        meta, _, _ = self.merge_metas_with_conflicts(api, dataset_id)

        multiple_items = self.items_count > 1
        datasets = []
        dataset_info = api.dataset.get_info_by_id(dataset_id)

        if multiple_items:
            logger.info(
                f"Found {self.items_count} topics in the input data."
                "Will create dataset in parent dataset for each topic."
            )
            nested_datasets = api.dataset.get_list(dataset_info.project_id, parent_id=dataset_id)
            existing_ds_names = set([ds.name for ds in nested_datasets])
            for item in self._items:
                ds_name = generate_free_name(existing_ds_names, get_file_name(item.path))
                ds = api.dataset.create(
                    dataset_info.project_id,
                    ds_name,
                    change_name_if_conflict=True,
                    parent_id=dataset_id,
                )
                existing_ds_names.add(ds.name)
                datasets.append(ds)

        if log_progress:
            progress, progress_cb = self.get_progress(self._total_msg_count, "Uploading...")
        else:
            progress_cb = None

        for idx, item in enumerate(self._items):
            current_dataset = dataset_info if not multiple_items else datasets[idx]
            current_dataset_id = current_dataset.id
            time_to_data = self.convert(item, meta)

            existing_names = set([pcd.name for pcd in api.pointcloud.get_list(current_dataset_id)])

            ann_episode = PointcloudEpisodeAnnotation()
            frame_to_pointcloud_ids = {}
            for idx, (time, data) in enumerate(time_to_data.items()):
                pcd_path = data["pcd"].as_posix()
                ann_path = data["ann"].as_posix() if data["ann"] is not None else None
                pcd_meta = data["meta"]

                pcd_name = generate_free_name(
                    existing_names, f"{time}.pcd", with_ext=True, extend_used_names=True
                )
                if is_episodes:
                    pcd_meta["frame"] = idx
                upload_fn = (
                    api.pointcloud_episode.upload_path
                    if is_episodes
                    else api.pointcloud.upload_path
                )
                info = upload_fn(current_dataset_id, pcd_name, pcd_path, pcd_meta)
                pcd_id = info.id
                frame_to_pointcloud_ids[idx] = pcd_id

                if ann_path is not None:
                    ann = PointcloudAnnotation.load_json_file(ann_path, meta)
                    if is_episodes:
                        objects = ann_episode.objects
                        figures = []
                        for fig in ann.figures:
                            obj_cls = meta.get_obj_class(fig.parent_object.obj_class.name)
                            if obj_cls is not None:
                                obj = PointcloudEpisodeObject(obj_cls)
                                objects = objects.add(obj)
                                figure = PointcloudFigure(obj, fig.geometry, frame_index=idx)
                                figures.append(figure)
                        frames = ann_episode.frames
                        frames = frames.add(PointcloudEpisodeFrame(idx, figures))
                        ann_episode = ann_episode.clone(objects=objects, frames=frames)
                    else:
                        api.pointcloud.annotation.append(pcd_id, ann)

                silent_remove(pcd_path)
                if ann_path is not None:
                    silent_remove(ann_path)
                if log_progress:
                    progress_cb(1)

            if is_episodes:
                ann_episode = ann_episode.clone(frames_count=len(time_to_data))
                api.pointcloud_episode.annotation.append(
                    current_dataset_id, ann_episode, frame_to_pointcloud_ids
                )

            logger.info(f"Dataset ID:{current_dataset_id} has been successfully uploaded.")

        if log_progress:
            if is_development():
                progress.close()



import random
import time
from typing import Any, Callable, Dict, List, Optional

import yaml

from supervisely._utils import is_development, logger
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.api.api import Api
from supervisely.api.video.video_api import VideoInfo
from supervisely.app.widgets import Button, Card, Container, Stepper, Widget
from supervisely.io import env
from supervisely.nn.inference.predict_app.gui.classes_selector import ClassesSelector
from supervisely.nn.inference.predict_app.gui.input_selector import InputSelector
from supervisely.nn.inference.predict_app.gui.model_selector import ModelSelector
from supervisely.nn.inference.predict_app.gui.output_selector import OutputSelector
from supervisely.nn.inference.predict_app.gui.preview import Preview
from supervisely.nn.inference.predict_app.gui.settings_selector import (
    AddPredictionsMode,
    SettingsSelector,
)
from supervisely.nn.inference.predict_app.gui.tags_selector import TagsSelector
from supervisely.nn.inference.predict_app.gui.utils import (
    copy_project,
    disable_enable,
    set_stepper_step,
    wrap_button_click,
)
from supervisely.nn.model.model_api import ModelAPI
from supervisely.nn.model.prediction import Prediction
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_annotation import VideoAnnotation


class StepFlow:

    def __init__(self, stepper: Stepper):
        self.stepper = stepper
        self.steps = {}
        self.step_sequence = []

    def register_step(
        self,
        name: str,
        card: Card,
        button: Optional[Button] = None,
        widgets_to_disable: Optional[List[Widget]] = None,
        validation_text: Optional[Widget] = None,
        validation_func: Optional[Callable] = None,
        position: Optional[int] = None,
    ) -> "StepFlow":
        self.steps[name] = {
            "card": card,
            "button": button,
            "widgets_to_disable": widgets_to_disable or [],
            "validation_text": validation_text,
            "validation_func": validation_func,
            "position": position,
            "next_steps": [],
            "on_select_click": [],
            "on_reselect_click": [],
            "wrapper": None,
            "has_button": button is not None,
        }

        if position is not None:
            while len(self.step_sequence) <= position:
                self.step_sequence.append(None)
            self.step_sequence[position] = name

        return self

    def set_next_steps(self, step_name: str, next_steps: List[str]) -> "StepFlow":
        if step_name in self.steps:
            self.steps[step_name]["next_steps"] = next_steps
        return self

    def add_on_select_actions(
        self, step_name: str, actions: List[Callable], is_reselect: bool = False
    ) -> "StepFlow":
        if step_name in self.steps:
            key = "on_reselect_click" if is_reselect else "on_select_click"
            self.steps[step_name][key].extend(actions)
        return self

    def build_wrappers(self) -> Dict[str, Callable]:
        valid_sequence = [s for s in self.step_sequence if s is not None and s in self.steps]

        for step_name in reversed(valid_sequence):
            step = self.steps[step_name]

            cards_to_unlock = []
            for next_step_name in step["next_steps"]:
                if next_step_name in self.steps:
                    cards_to_unlock.append(self.steps[next_step_name]["card"])

            callback = None
            if step["next_steps"] and step["has_button"]:
                for next_step_name in step["next_steps"]:
                    if (
                        next_step_name in self.steps
                        and self.steps[next_step_name].get("wrapper")
                        and self.steps[next_step_name]["has_button"]
                    ):
                        callback = self.steps[next_step_name]["wrapper"]
                        break

            if step["has_button"]:
                wrapper = wrap_button_click(
                    button=step["button"],
                    cards_to_unlock=cards_to_unlock,
                    widgets_to_disable=step["widgets_to_disable"],
                    callback=callback,
                    validation_text=step["validation_text"],
                    validation_func=step["validation_func"],
                    on_select_click=step["on_select_click"],
                    on_reselect_click=step["on_reselect_click"],
                    collapse_card=None,
                )

                step["wrapper"] = wrapper

        return {
            name: self.steps[name]["wrapper"]
            for name in self.steps
            if self.steps[name].get("wrapper") and self.steps[name]["has_button"]
        }

    def setup_button_handlers(self) -> None:
        positions = {}
        pos = 1

        for i, step_name in enumerate(self.step_sequence):
            if step_name is not None and step_name in self.steps:
                positions[step_name] = pos
                pos += 1

        for step_name, step in self.steps.items():
            if step_name in positions and step.get("wrapper") and step["has_button"]:

                button = step["button"]
                wrapper = step["wrapper"]
                position = positions[step_name]
                next_position = position + 1

                def create_handler(btn, cb, next_pos):
                    def handler():
                        cb()
                        set_stepper_step(self.stepper, btn, next_pos=next_pos)

                    return handler

                button.click(create_handler(button, wrapper, next_position))

    def build(self) -> Dict[str, Callable]:
        wrappers = self.build_wrappers()
        self.setup_button_handlers()
        return wrappers


class PredictAppGui:
    def __init__(self, api: Api, static_dir: str = "static"):
        self.api = api
        self.static_dir = static_dir

        # Environment variables
        self.team_id = env.team_id()
        self.workspace_id = env.workspace_id()
        self.project_id = env.project_id(raise_not_found=False)
        # -------------------------------- #

        # Flags
        self._stop_flag = False
        self._is_running = False
        # -------------------------------- #

        # GUI
        # Steps
        self.steps = []

        # 1. Input selector
        self.input_selector = InputSelector(self.workspace_id)
        self.steps.append(self.input_selector.card)

        # 2. Model selector
        self.model_selector = ModelSelector(self.api, self.team_id)
        self.steps.append(self.model_selector.card)

        # 3. Classes selector
        self.classes_selector = None
        if True:
            self.classes_selector = ClassesSelector()
            self.steps.append(self.classes_selector.card)

        # 4. Tags selector
        self.tags_selector = None
        if False:
            self.tags_selector = TagsSelector()
            self.steps.append(self.tags_selector.card)

        # 5. Settings selector
        self.settings_selector = SettingsSelector()
        self.steps.append(self.settings_selector.card)

        # 6. Preview
        self.preview = None
        if False:
            self.preview = Preview(api, static_dir)
            self.steps.append(self.preview.card)

        # 7. Output selector
        self.output_selector = OutputSelector(self.api)
        self.steps.append(self.output_selector.card)
        # -------------------------------- #

        # Stepper
        self.stepper = Stepper(widgets=self.steps)
        # ---------------------------- #

        # Layout
        self.layout = Container([self.stepper])
        # ---------------------------- #

        # Button Utils
        def deploy_model() -> ModelAPI:
            self.model_selector.validator_text.hide()
            model_api = None
            try:
                model_api = type(self.model_selector.model).deploy(self.model_selector.model)
            except:
                self.output_selector.start_button.disable()
                raise
            else:
                self.output_selector.start_button.enable()
            return model_api

        # Reimplement deploy method for DeployModel widget
        self.model_selector.model.deploy = deploy_model

        def set_entity_meta():
            model_api = self.model_selector.model.model_api

            model_meta = model_api.get_model_meta()
            if self.classes_selector is not None:
                self.classes_selector.classes_table.set_project_meta(model_meta)
                self.classes_selector.classes_table.show()
            if self.tags_selector is not None:
                self.tags_selector.tags_table.set_project_meta(model_meta)
                self.tags_selector.tags_table.show()

            inference_settings = model_api.get_settings()
            self.settings_selector.set_inference_settings(inference_settings)

            if self.preview is not None:
                self.preview.inference_settings = inference_settings

        def reset_entity_meta():
            empty_meta = ProjectMeta()
            if self.classes_selector is not None:
                self.classes_selector.classes_table.set_project_meta(empty_meta)
                self.classes_selector.classes_table.hide()
            if self.tags_selector is not None:
                self.tags_selector.tags_table.set_project_meta(empty_meta)
                self.tags_selector.tags_table.hide()

            self.settings_selector.set_inference_settings("")

            if self.preview is not None:
                self.preview.inference_settings = None

        def disable_settings_editor():
            if self.settings_selector.inference_settings.readonly:
                self.settings_selector.inference_settings.readonly = False
            else:
                self.settings_selector.inference_settings.readonly = True

        def generate_preview():
            def _get_frame_annotation(
                video_info: VideoInfo, frame_index: int, project_meta: ProjectMeta
            ) -> Annotation:
                video_annotation = VideoAnnotation.from_json(
                    self.api.video.annotation.download(video_info.id, frame_index),
                    project_meta=project_meta,
                    key_id_map=KeyIdMap(),
                )
                frame = video_annotation.frames.get(frame_index)
                img_size = (video_info.frame_height, video_info.frame_width)
                if frame is None:
                    return Annotation(img_size)
                labels = []
                for figure in frame.figures:
                    labels.append(Label(figure.geometry, figure.video_object.obj_class))
                ann = Annotation(img_size, labels=labels)
                return ann

            if self.preview is None:
                return

            self.preview.validator_text.hide()
            self.preview.gallery.clean_up()
            self.preview.gallery.show()
            self.preview.gallery.loading = True
            try:
                items_settings = self.input_selector.get_settings()
                if "video_id" in items_settings:
                    video_id = items_settings["video_id"]
                    video_info = self.api.video.get_info_by_id(video_id)
                    video_frame = random.randint(0, video_info.frames_count - 1)
                    self.api.video.frame.download_path(
                        video_info.id, video_frame, self.preview.preview_path
                    )
                    img_url = self.preview.peview_url
                    project_meta = ProjectMeta.from_json(
                        self.api.project.get_meta(video_info.project_id)
                    )
                    input_ann = _get_frame_annotation(video_info, video_frame, project_meta)
                    prediction = self.model_selector.model.model_api.predict(
                        input=self.preview.preview_path, **self.settings_selector.get_settings()
                    )[0]
                    output_ann = prediction.annotation
                else:
                    if "project_id" in items_settings:
                        project_id = items_settings["project_id"]
                        dataset_infos = self.api.dataset.get_list(project_id, recursive=True)
                        dataset_infos = [ds for ds in dataset_infos if ds.items_count > 0]
                        if not dataset_infos:
                            raise ValueError("No datasets with items found in the project.")
                        dataset_info = random.choice(dataset_infos)
                    elif "dataset_ids" in items_settings:
                        dataset_ids = items_settings["dataset_ids"]
                        dataset_infos = [
                            self.api.dataset.get_info_by_id(dataset_id)
                            for dataset_id in dataset_ids
                        ]
                        dataset_infos = [ds for ds in dataset_infos if ds.items_count > 0]
                        if not dataset_infos:
                            raise ValueError("No items in selected datasets.")
                        dataset_info = random.choice(dataset_infos)
                    else:
                        raise ValueError("No valid item settings found for preview.")
                    images = self.api.image.get_list(dataset_info.id)
                    image_info = random.choice(images)
                    img_url = image_info.preview_url

                    project_meta = ProjectMeta.from_json(
                        self.api.project.get_meta(dataset_info.project_id)
                    )
                    input_ann = Annotation.from_json(
                        self.api.annotation.download(image_info.id).annotation,
                        project_meta=project_meta,
                    )
                    prediction = self.model_selector.model.model_api.predict(
                        image_id=image_info.id, **self.settings_selector.get_settings()
                    )[0]
                    output_ann = prediction.annotation

                self.preview.gallery.append(img_url, input_ann, "Input")
                self.preview.gallery.append(img_url, output_ann, "Output")
                self.preview.validator_text.hide()
                self.preview.gallery.show()
                return prediction
            except Exception as e:
                self.preview.gallery.hide()
                self.preview.validator_text.set(
                    text=f"Error during preview: {str(e)}", status="error"
                )
                self.preview.validator_text.show()
                self.preview.gallery.clean_up()
            finally:
                self.preview.gallery.loading = False

        # ---------------------------- #

        # StepFlow callbacks and wiring
        self.step_flow = StepFlow(self.stepper)
        position = 0

        # 1. Input selector
        self.step_flow.register_step(
            "input_selector",
            self.input_selector.card,
            self.input_selector.button,
            self.input_selector.widgets_to_disable,
            self.input_selector.validator_text,
            self.input_selector.validate_step,
            position=position,
        )
        position += 1

        # 2. Model selector
        self.step_flow.register_step(
            "model_selector",
            self.model_selector.card,
            self.model_selector.button,
            self.model_selector.widgets_to_disable,
            self.model_selector.validator_text,
            self.model_selector.validate_step,
            position=position,
        )
        self.step_flow.add_on_select_actions("model_selector", [set_entity_meta])
        self.step_flow.add_on_select_actions(
            "model_selector", [reset_entity_meta], is_reselect=True
        )
        position += 1

        # 3. Classes selector
        if self.classes_selector is not None:
            self.step_flow.register_step(
                "classes_selector",
                self.classes_selector.card,
                self.classes_selector.button,
                self.classes_selector.widgets_to_disable,
                self.classes_selector.validator_text,
                self.classes_selector.validate_step,
                position=position,
            )
            position += 1

        # 4. Tags selector
        if self.tags_selector is not None:
            self.step_flow.register_step(
                "tags_selector",
                self.tags_selector.card,
                self.tags_selector.button,
                self.tags_selector.widgets_to_disable,
                self.tags_selector.validator_text,
                self.tags_selector.validate_step,
                position=position,
            )
            position += 1

        # 5. Settings selector
        self.step_flow.register_step(
            "settings_selector",
            self.settings_selector.card,
            self.settings_selector.button,
            self.settings_selector.widgets_to_disable,
            self.settings_selector.validator_text,
            self.settings_selector.validate_step,
            position=position,
        )
        self.step_flow.add_on_select_actions("settings_selector", [disable_settings_editor])
        self.step_flow.add_on_select_actions("settings_selector", [disable_settings_editor], True)
        position += 1

        # 6. Preview
        if self.preview is not None:
            self.step_flow.register_step(
                "preview",
                self.preview.card,
                self.preview.button,
                self.preview.widgets_to_disable,
                self.preview.validator_text,
                self.preview.validate_step,
                position=position,
            ).add_on_select_actions("preview", [generate_preview])
            position += 1

        # 7. Output selector
        self.step_flow.register_step(
            "output_selector",
            self.output_selector.card,
            None,
            self.output_selector.widgets_to_disable,
            self.output_selector.validator_text,
            self.output_selector.validate_step,
            position=position,
        )

        # Dependencies Chain
        has_model_selector = self.model_selector is not None
        has_classes_selector = self.classes_selector is not None
        has_tags_selector = self.tags_selector is not None
        has_preview = self.preview is not None

        # Step 1 -> Step 2
        prev_step = "input_selector"
        if has_model_selector:
            self.step_flow.set_next_steps(prev_step, ["model_selector"])
            prev_step = "model_selector"
        # Step 2 -> Step 3
        if has_classes_selector:
            self.step_flow.set_next_steps(prev_step, ["classes_selector"])
            prev_step = "classes_selector"
        # Step 3 -> Step 4
        if has_tags_selector:
            self.step_flow.set_next_steps(prev_step, ["tags_selector"])
            prev_step = "tags_selector"
        # Step 4 -> Step 5
        self.step_flow.set_next_steps(prev_step, ["settings_selector"])
        prev_step = "settings_selector"
        # Step 5 -> Step 6
        if has_preview:
            self.step_flow.set_next_steps(prev_step, ["preview"])
            prev_step = "preview"
        # Step 6 -> Step 7
        self.step_flow.set_next_steps(prev_step, ["output_selector"])

        # Create all wrappers and set button handlers
        wrappers = self.step_flow.build()

        self.input_selector_cb = wrappers.get("input_selector")
        self.classes_selector_cb = wrappers.get("classes_selector")
        self.tags_selector_cb = wrappers.get("tags_selector")
        self.model_selector_cb = wrappers.get("model_selector")
        self.settings_selector_cb = wrappers.get("settings_selector")
        self.preview_cb = wrappers.get("preview")
        self.output_selector_cb = wrappers.get("output_selector")
        # ------------------------------------------------- #

        # Other Handlers
        @self.input_selector.radio.value_changed
        def input_selector_type_changed(value: str):
            self.input_selector.validator_text.hide()

        @self.input_selector.select_dataset_for_video.value_changed
        def dataset_for_video_changed(dataset_id: int):
            self.input_selector.select_video.loading = True
            if dataset_id is None:
                rows = []
            else:
                dataset_info = self.api.dataset.get_info_by_id(dataset_id)
                videos = self.api.video.get_list(dataset_id)
                rows = [[video.id, video.name, dataset_info.name] for video in videos]
            self.input_selector.select_video.rows = rows
            self.input_selector.select_video.loading = False

        # ------------------------------------------------- #

    def run(self, run_parameters: Dict[str, Any] = None) -> List[Prediction]:
        self.show_validator_text()
        self.set_validator_text("Preparing settings for prediction...", "info")
        if run_parameters is None:
            run_parameters = self.get_run_parameters()

        if self.model_selector.model.model_api is None:
            self.model_selector.model._deploy()

        model_api = self.model_selector.model.model_api
        if model_api is None:
            logger.error("Model Deployed with an error")
            self.set_validator_text("Model Deployed with an error", "error")
            return

        kwargs = {}

        # Input
        # Input would be newely created project
        input_parameters = run_parameters["input"]
        input_project_id = input_parameters.get("project_id", None)
        if input_project_id is None:
            raise ValueError("Input project ID is required for prediction.")
        input_dataset_ids = input_parameters.get("dataset_ids", [])
        if not input_dataset_ids:
            raise ValueError("At least one dataset must be selected for prediction.")

        # Settings
        settings = run_parameters["settings"]
        prediction_mode = settings.pop("predictions_mode")
        upload_mode = None
        with_annotations = None
        if prediction_mode == AddPredictionsMode.REPLACE_EXISTING_LABELS:
            upload_mode = "replace"
            with_annotations = False
        elif prediction_mode == AddPredictionsMode.MERGE_WITH_EXISTING_LABELS:
            upload_mode = "append"
            with_annotations = True
        elif prediction_mode == AddPredictionsMode.REPLACE_EXISTING_LABELS_AND_SAVE_IMAGE_TAGS:
            upload_mode = "replace"
            with_annotations = True
        kwargs.update(settings)
        kwargs["upload_mode"] = upload_mode

        # Classes
        classes = run_parameters["classes"]
        if classes:
            kwargs["classes"] = classes

        # Output
        # Always create new project
        # But the actual inference will happen inplace
        output_parameters = run_parameters["output"]
        project_name = output_parameters["project_name"]
        if not project_name:
            input_project_info = self.api.project.get_info_by_id(input_project_id)
            project_name = input_project_info.name + " [Predictions]"
            logger.warning("Project name is empty, using auto-generated name: " + project_name)

        # Copy project
        self.set_validator_text("Copying project...", "info")
        created_project = copy_project(
            self.api,
            project_name,
            self.workspace_id,
            input_project_id,
            input_dataset_ids,
            with_annotations,
            self.output_selector.progress,
        )
        # ------------------------ #

        # Run prediction
        self.set_validator_text("Running prediction...", "info")
        predictions = []
        self._is_running = True
        try:
            with model_api.predict_detached(
                project_id=created_project.id,
                tqdm=self.output_selector.progress(),
                **kwargs,
            ) as session:
                self.output_selector.progress.show()
                i = 0
                for prediction in session:
                    predictions.append(prediction)
                    i += 1
                    if self._stop_flag:
                        logger.info("Prediction stopped by user.")
                        break
            self.output_selector.progress.hide()
        except Exception as e:
            self.output_selector.progress.hide()
            logger.error(f"Error during prediction: {str(e)}")
            self.set_validator_text(f"Error during prediction: {str(e)}", "error")
            disable_enable(self.output_selector.widgets_to_disable, False)
            self._is_running = False
            self._stop_flag = False
            raise e
        finally:
            self._is_running = False
            self._stop_flag = False
        # ------------------------ #

        # Set result thumbnail
        self.set_validator_text("Project successfully processed", "success")
        self.output_selector.set_result_thumbnail(created_project.id)
        # ------------------------ #
        return predictions

    def stop(self):
        logger.info("Stopping prediction...")
        self._stop_flag = True

    def wait_for_stop(self, timeout: int = None):
        logger.info(
            "Waiting " + ""
            if timeout is None
            else f"{timeout} seconds " + "for prediction to stop..."
        )
        t = time.monotonic()
        while self._is_running:
            if timeout is not None and time.monotonic() - t > timeout:
                raise TimeoutError("Timeout while waiting for stop.")
            time.sleep(0.1)
        logger.info("Prediction stopped.")

    def shutdown_model(self):
        self.stop()
        self.wait_for_stop(10)
        self.model_selector.model.stop()

    def get_run_parameters(self) -> Dict[str, Any]:
        settings = {
            "model": self.model_selector.model.get_deploy_parameters(),
            "settings": self.settings_selector.get_settings(),
            "input": self.input_selector.get_settings(),
            "output": self.output_selector.get_settings(),
        }
        if self.classes_selector is not None:
            settings["classes"] = self.classes_selector.get_selected_classes()
        if self.tags_selector is not None:
            settings["tags"] = self.tags_selector.get_selected_tags()
        return settings

    def load_from_json(self, data):
        # 1. Input selector
        self.input_selector.load_from_json(data.get("input", {}))
        # self.input_selector_cb()

        # 2. Model selector
        self.model_selector.model.load_from_json(data.get("model", {}))

        # 3. Classes selector
        if self.classes_selector is not None:
            self.classes_selector.load_from_json(data.get("classes", {}))

        # 4. Tags selector
        if self.tags_selector is not None:
            self.tags_selector.load_from_json(data.get("tags", {}))

        # 5. Settings selector
        self.settings_selector.load_from_json(data.get("settings", {}))

        # 6. Preview (No need?)
        if self.preview is not None:
            self.preview.load_from_json(data.get("preview", {}))

        # 7. Output selector
        self.output_selector.load_from_json(data.get("output", {}))

    def set_validator_text(self, text: str, status: str = "text"):
        self.output_selector.validator_text.set(text=text, status=status)

    def show_validator_text(self):
        self.output_selector.validator_text.show()

    def hide_validator_text(self):
        self.output_selector.validator_text.hide()



from typing import List, Tuple

from supervisely.annotation.annotation import Annotation
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import ApiField
from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import GalleryWidget, MarkdownWidget
from supervisely.project.project_meta import ProjectMeta


class ExplorePredictions(BaseVisMetrics):

    MARKDOWN_DIFFERENCE = "markdown_explore_difference"
    GALLERY_DIFFERENCE = "explore_difference_gallery"
    MARKDOWN_SAME_ERRORS = "markdown_explore_same_errors"
    GALLERY_SAME_ERRORS = "explore_same_error_gallery"

    @property
    def difference_predictions_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_explore_difference
        return MarkdownWidget(self.MARKDOWN_DIFFERENCE, "Explore Predictions", text)

    @property
    def explore_gallery(self) -> GalleryWidget:
        columns_number = len(self.eval_results) + 1
        *data, min_conf = self._get_sample_data()
        default_filters = [{"confidence": [min_conf, 1]}]
        gallery = GalleryWidget(
            self.GALLERY_DIFFERENCE, columns_number=columns_number, filters=default_filters
        )
        gallery.add_image_left_header("Click to explore more")
        gallery.set_project_meta(self.eval_results[0].gt_project_meta)
        gallery.set_images(*data)
        click_data = self.get_click_data_explore_all()
        gallery.set_click_data(self.explore_modal_table.id, click_data)
        gallery.set_show_all_data(self.explore_modal_table.id, click_data)
        gallery._gallery._update_filters()

        return gallery

    def _get_sample_data(self) -> Tuple[List[ImageInfo], List[Annotation], List[ProjectMeta]]:
        images = []
        annotations = []
        metas = [self.eval_results[0].gt_project_meta]
        skip_tags_filtering = []
        api = self.eval_results[0].api
        min_conf = float("inf")
        names = None
        ds_name = None
        for idx, eval_res in enumerate(self.eval_results):
            if idx == 0:
                dataset_info = eval_res.gt_dataset_infos[0]
                image_infos = api.image.get_list(dataset_info.id, limit=5, force_metadata_for_links=False)
                ds_name = dataset_info.name
                images_ids = [image_info.id for image_info in image_infos]
                names = [image_info.name for image_info in image_infos]
                images.append(image_infos)
                anns = api.annotation.download_batch(dataset_info.id, images_ids, force_metadata_for_links=False)
                annotations.append(anns)
                skip_tags_filtering.append(True)
            metas.append(eval_res.pred_project_meta)
            assert ds_name is not None, "Failed to get GT dataset name for gallery"

            dataset_info = api.dataset.get_info_by_name(eval_res.pred_project_id, ds_name)

            assert names is not None, "Failed to get GT image names for gallery"
            image_infos = eval_res.api.image.get_list(
                dataset_info.id,
                filters=[
                    {ApiField.FIELD: ApiField.NAME, ApiField.OPERATOR: "in", ApiField.VALUE: names}
                ],
                force_metadata_for_links=False,
            )
            images_ids = [image_info.id for image_info in image_infos]
            images.append(image_infos)
            anns = eval_res.api.annotation.download_batch(dataset_info.id, images_ids, force_metadata_for_links=False)
            annotations.append(anns)
            skip_tags_filtering.append(False)
            min_conf = min(min_conf, eval_res.mp.conf_threshold)

        images = list(i for x in zip(*images) for i in x)
        annotations = list(i for x in zip(*annotations) for i in x)
        return images, annotations, metas, skip_tags_filtering, min_conf

    def get_click_data_explore_all(self) -> dict:
        res = {}

        res["projectMeta"] = self.eval_results[0].gt_project_meta.to_json()
        res["layoutTemplate"] = [None, None, None]

        res["layoutTemplate"] = [{"skipObjectTagsFiltering": True, "columnTitle": "Ground Truth"}]
        # for i in range(len(self.eval_results)):
        for idx, eval_res in enumerate(self.eval_results, 1):
            res["layoutTemplate"].append({"columnTitle": f"[{idx}] {eval_res.name}"})

        click_data = res.setdefault("clickData", {})
        explore = click_data.setdefault("explore", {})
        explore["title"] = "Explore all predictions"

        images_ids = []
        api = self.eval_results[0].api
        min_conf = float("inf")
        names = None
        ds_names = None
        for idx, eval_res in enumerate(self.eval_results):
            if idx == 0:
                dataset_infos = eval_res.gt_dataset_infos
                ds_names = [ds.name for ds in dataset_infos]
                current_images_ids = []
                current_images_names = []
                for ds in dataset_infos:
                    image_infos = api.image.get_list(ds.id, force_metadata_for_links=False)
                    image_infos = sorted(image_infos, key=lambda x: x.name)
                    current_images_names.extend([image_info.name for image_info in image_infos])
                    current_images_ids.extend([image_info.id for image_info in image_infos])
                images_ids.append(current_images_ids)
                names = current_images_names

            dataset_infos = api.dataset.get_list(eval_res.pred_project_id)
            dataset_infos = [ds for ds in dataset_infos if ds.name in ds_names]
            dataset_infos = sorted(dataset_infos, key=lambda x: ds_names.index(x.name))
            current_images_infos = []
            for ds in dataset_infos:
                image_infos = api.image.get_list(ds.id, force_metadata_for_links=False)
                image_infos = [image_info for image_info in image_infos if image_info.name in names]
                current_images_infos.extend(image_infos)
            current_images_infos = sorted(current_images_infos, key=lambda x: names.index(x.name))
            images_ids.append([image_info.id for image_info in current_images_infos])

            min_conf = min(min_conf, eval_res.mp.conf_threshold)

        explore["imagesIds"] = list(i for x in zip(*images_ids) for i in x)
        explore["filters"] = [{"type": "tag", "tagId": "confidence", "value": [min_conf, 1]}]

        return res



from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
    TableWidget,
)


class CalibrationScore(BaseVisMetrics):
    @property
    def header_md(self) -> MarkdownWidget:
        text_template = self.vis_texts.markdown_calibration_score_1
        text = text_template.format(self.vis_texts.definitions.confidence_score)
        return MarkdownWidget(
            name="markdown_calibration_score",
            title="Calibration Score",
            text=text,
        )

    @property
    def collapse_tip(self) -> CollapseWidget:
        md = MarkdownWidget(
            name="what_is_calibration",
            title="What is calibration?",
            text=self.vis_texts.markdown_what_is_calibration,
        )
        return CollapseWidget([md])

    @property
    def header_md_2(self) -> MarkdownWidget:
        return MarkdownWidget(
            name="markdown_calibration_score_2",
            title="",
            text=self.vis_texts.markdown_calibration_score_2,
        )

    @property
    def table(self) -> TableWidget:
        columns = [" ", "confidence threshold", "ECE", "MCE"]
        columns_options = [
            {"disableSort": True},
            {"disableSort": True},
            {"disableSort": True},
            {"disableSort": True},
        ]
        content = []
        for i, eval_result in enumerate(self.eval_results):
            name = f"[{i+1}] {eval_result.name}"
            conf_threshold = eval_result.mp.m_full.get_f1_optimal_conf()[0] or 0.0
            ece = eval_result.mp.m_full.expected_calibration_error()
            mce = eval_result.mp.m_full.maximum_calibration_error()
            row = [name, round(conf_threshold, 2), round(ece, 2), round(mce, 2)]
            dct = {
                "row": row,
                "id": name,
                "items": row,
            }
            content.append(dct)
        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        return TableWidget(
            "table_reliability",
            data,
            show_header_controls=False,
            # main_column=columns[0],
            fix_columns=1,
        )

    @property
    def reliability_diagram_md(self) -> MarkdownWidget:
        return MarkdownWidget(
            name="markdown_reliability_diagram",
            title="Reliability Diagram",
            text=self.vis_texts.markdown_reliability_diagram,
        )

    @property
    def reliability_chart(self) -> ChartWidget:
        return ChartWidget(name="chart_reliability", figure=self.get_rel_figure())

    @property
    def collapse_ece(self) -> CollapseWidget:
        md = MarkdownWidget(
            name="markdown_calibration_curve_interpretation",
            title="How to interpret the Calibration curve",
            text=self.vis_texts.markdown_calibration_curve_interpretation,
        )
        return CollapseWidget([md])

    @property
    def confidence_score_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_confidence_score_1.format(
            self.vis_texts.definitions.confidence_threshold
        )
        return MarkdownWidget(
            "markdown_confidence_score_1",
            "Confidence Score Profile",
            text,
        )

    @property
    def confidence_chart(self) -> ChartWidget:
        return ChartWidget(name="chart_confidence", figure=self.get_conf_figure())

    @property
    def confidence_score_md_2(self) -> MarkdownWidget:
        return MarkdownWidget(
            name="markdown_confidence_score_2",
            title="",
            text=self.vis_texts.markdown_confidence_score_2,
        )

    @property
    def collapse_conf_score(self) -> CollapseWidget:
        md = MarkdownWidget(
            name="markdown_plot_confidence_profile",
            title="How to plot Confidence Profile?",
            text=self.vis_texts.markdown_plot_confidence_profile,
        )
        return CollapseWidget([md])

    def get_conf_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        # Create an empty figure
        fig = go.Figure()

        for i, eval_result in enumerate(self.eval_results):
            # Add a line trace for each eval_result
            fig.add_trace(
                go.Scatter(
                    x=eval_result.dfsp_down["scores"],
                    y=eval_result.dfsp_down["f1"],
                    mode="lines",
                    name=f"[{i+1}] {eval_result.name}",
                    line=dict(color=eval_result.color),
                    hovertemplate="Confidence Score: %{x:.2f}<br>Value: %{y:.2f}<extra></extra>",
                )
            )

            # Add a vertical line and annotation for F1-optimal threshold if available
            if eval_result.mp.f1_optimal_conf is not None and eval_result.mp.best_f1 is not None:
                fig.add_shape(
                    type="line",
                    x0=eval_result.mp.f1_optimal_conf,
                    x1=eval_result.mp.f1_optimal_conf,
                    y0=0,
                    y1=eval_result.mp.best_f1,
                    line=dict(color="gray", width=2, dash="dash"),
                    name=f"F1-optimal threshold ({eval_result.name})",
                )
            if eval_result.mp.custom_conf_threshold is not None:
                fig.add_shape(
                    type="line",
                    x0=eval_result.mp.custom_conf_threshold,
                    x1=eval_result.mp.custom_conf_threshold,
                    y0=0,
                    y1=eval_result.mp.custom_f1,
                    line=dict(color="black", width=2, dash="dash"),
                    name=f"Confidence threshold ({eval_result.name})",
                )

        # Update the layout
        fig.update_layout(
            yaxis=dict(range=[0, 1], title="Scores"),
            xaxis=dict(range=[0, 1], tick0=0, dtick=0.1, title="Confidence Score"),
            height=500,
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
            showlegend=True,  # Show legend to differentiate between results
        )

        return fig

    def get_rel_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        for i, eval_result in enumerate(self.eval_results):
            # Calibration curve (only positive predictions)
            true_probs, pred_probs = eval_result.mp.m_full.calibration_curve()

            fig.add_trace(
                go.Scatter(
                    x=pred_probs,
                    y=true_probs,
                    mode="lines+markers",
                    name=f"[{i+1}] {eval_result.name}",
                    line=dict(color=eval_result.color),
                    hovertemplate=f"{eval_result.name}<br>"
                    + "Confidence Score: %{x:.2f}<br>Fraction of True Positives: %{y:.2f}<extra></extra>",
                )
            )

        fig.add_trace(
            go.Scatter(
                x=[0, 1],
                y=[0, 1],
                mode="lines",
                name="Perfectly calibrated",
                line=dict(color="orange", dash="dash"),
            )
        )

        fig.update_layout(
            # title="Calibration Curve (only positive predictions)",
            xaxis_title="Confidence Score",
            yaxis_title="Fraction of True Positives",
            legend=dict(x=0.6, y=0.1),
            xaxis=dict(range=[0, 1.1]),
            yaxis=dict(range=[0, 1.1]),
            width=700,
            height=500,
        )
        return fig



import json
import time
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterator, List, Literal, Tuple, Union

import numpy as np
import requests
from requests import Timeout
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm.auto import tqdm

import supervisely.io.env as env
from supervisely._utils import get_valid_kwargs, logger
from supervisely.api.api import Api
from supervisely.imaging._video import ALLOWED_VIDEO_EXTENSIONS
from supervisely.imaging.image import SUPPORTED_IMG_EXTS, write_bytes
from supervisely.io.fs import (
    dir_exists,
    file_exists,
    get_file_ext,
    get_file_size,
    list_files,
    list_files_recursively,
)
from supervisely.io.network_exceptions import process_requests_exception
from supervisely.nn.model.prediction import Prediction
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta


def value_generator(value):
    while True:
        yield value


class PredictionSession:

    class Iterator:
        def __init__(self, total, session: "PredictionSession", tqdm: tqdm = None):
            self.total = total
            self.session = session
            self.results_queue = []
            self.tqdm = tqdm

        def __len__(self) -> int:
            return self.total

        def __iter__(self) -> Iterator:
            return self

        def __next__(self) -> Dict[str, Any]:
            if not self.results_queue:
                pending_results = self.session._wait_for_pending_results(tqdm=self.tqdm)
                self.results_queue += pending_results
            if not self.results_queue:
                raise StopIteration
            pred = self.results_queue.pop(0)
            return pred

    def __init__(
        self,
        url: str,
        input: Union[np.ndarray, str, PathLike, list] = None,
        image_id: Union[List[int], int] = None,
        video_id: Union[List[int], int] = None,
        dataset_id: Union[List[int], int] = None,
        project_id: Union[List[int], int] = None,
        api: "Api" = None,
        tracking: bool = None,
        tracking_config: dict = None,
        **kwargs: dict,
    ): 
                  
        extra_input_args = ["image_ids", "video_ids", "dataset_ids", "project_ids"]
        assert (
            sum(
                [
                    x is not None
                    for x in [
                        input,
                        image_id,
                        video_id,
                        dataset_id,
                        project_id,
                        *[kwargs.get(extra_input, None) for extra_input in extra_input_args],
                    ]
                ]
            )
            == 1
        ), "Exactly one of input, image_ids, video_id, dataset_id, project_id or image_id must be provided."

        
        self._iterator = None
        self._base_url = url
        self.inference_request_uuid = None
        self.input = input
        self.api = api

        self.api_token = self._get_api_token()
        self._model_meta = None
        self.final_result = None

        if "stride" in kwargs:
            kwargs["step"] = kwargs["stride"]
        if "start_frame" in kwargs:
            kwargs["start_frame_index"] = kwargs["start_frame"]
        if "num_frames" in kwargs:
            kwargs["frames_count"] = kwargs["num_frames"]
        self.kwargs = kwargs
        if kwargs.get("show_progress", False) and "tqdm" not in kwargs:
            kwargs["tqdm"] = tqdm()
        self.tqdm = kwargs.pop("tqdm", None)

        self.inference_settings = {
            k: v for k, v in kwargs.items() if isinstance(v, (str, int, float))
        }
        
        if tracking is True:
            model_info = self._get_session_info()
            if not model_info.get("tracking_on_videos_support", False):
                raise ValueError("Tracking is not supported by this model")
            
            if tracking_config is None:
                self.tracker = "botsort"
                self.tracker_settings = {}
            else:
                cfg = dict(tracking_config)
                self.tracker = cfg.pop("tracker", "botsort")
                self.tracker_settings = cfg
        else:
            self.tracker = None
            self.tracker_settings = None

        if "classes" in kwargs:
            self.inference_settings["classes"] = kwargs["classes"]
        # TODO: remove "settings", it is the same as inference_settings
        if "settings" in kwargs:
            self.inference_settings.update(kwargs["settings"])
        if "inference_settings" in kwargs:
            self.inference_settings.update(kwargs["inference_settings"])

        # extra input args
        image_ids = self._set_var_from_kwargs("image_ids", kwargs, image_id)
        video_ids = self._set_var_from_kwargs("video_ids", kwargs, video_id)
        dataset_ids = self._set_var_from_kwargs("dataset_ids", kwargs, dataset_id)
        project_ids = self._set_var_from_kwargs("project_ids", kwargs, project_id)
        source = next(
            x
            for x in [
                input,
                image_id,
                video_id,
                dataset_id,
                project_id,
                image_ids,
                video_ids,
                dataset_ids,
                project_ids,
            ]
            if x is not None
        )
        self.kwargs["source"] = source
        self.prediction_kwargs_iterator = value_generator({})

        if not isinstance(input, list):
            input = [input]
        if isinstance(input[0], np.ndarray):
            # input is numpy array
            self._predict_images(input, **kwargs)
        elif isinstance(input[0], (str, PathLike)):
            if len(input) > 1:
                # if the input is a list of paths, assume they are images
                for x in input:
                    if not isinstance(x, (str, PathLike)):
                        raise ValueError("Input must be a list of strings or PathLike objects.")
                self._iterator = self._predict_images_bytes(input, **kwargs)
            else:
                if dir_exists(input[0]):
                    try:
                        project = Project(str(input[0]), mode=OpenMode.READ)
                    except Exception:
                        project = None
                    image_paths = []
                    if project is not None:
                        for dataset in project.datasets:
                            dataset: Dataset
                            for _, image_path, _ in dataset.items():
                                image_paths.append(image_path)
                    else:
                        # if the input is a directory, assume it contains images
                        recursive = kwargs.get("recursive", False)
                        if recursive:
                            image_paths = list_files_recursively(
                                input[0], valid_extensions=SUPPORTED_IMG_EXTS
                            )
                        else:
                            image_paths = list_files(input[0], valid_extensions=SUPPORTED_IMG_EXTS)
                    if len(image_paths) == 0:
                        raise ValueError("Directory is empty.")
                    self._iterator = self._predict_images(image_paths, **kwargs)
                elif file_exists(input[0]):
                    ext = get_file_ext(input[0])
                    if ext == "":
                        raise ValueError("File has no extension.")
                    if ext.lower() in SUPPORTED_IMG_EXTS:
                        self._iterator = self._predict_images(input, **kwargs)
                    elif ext.lower() in ALLOWED_VIDEO_EXTENSIONS:
                        kwargs = get_valid_kwargs(kwargs, self._predict_videos, exclude=["videos"])
                        self._iterator = self._predict_videos(input, tracker=self.tracker, tracker_settings=self.tracker_settings, **kwargs)
                    else:
                        raise ValueError(
                            f"Unsupported file extension: {ext}. Supported extensions are: {SUPPORTED_IMG_EXTS + ALLOWED_VIDEO_EXTENSIONS}"
                        )
                else:
                    raise ValueError(f"File or directory does not exist: {input[0]}")
        elif image_ids is not None:
            self._iterator = self._predict_images(image_ids, **kwargs)
        elif video_ids is not None:
            if len(video_ids) > 1:
                raise ValueError("Only one video id can be provided.")
            kwargs = get_valid_kwargs(kwargs, self._predict_videos, exclude=["videos"])
            self._iterator = self._predict_videos(video_ids, tracker=self.tracker, tracker_settings=self.tracker_settings, **kwargs)
        elif dataset_ids is not None:
            kwargs = get_valid_kwargs(
                kwargs,
                self._predict_datasets,
                exclude=["dataset_ids"],
            )
            self._iterator = self._predict_datasets(dataset_ids, **kwargs)
        elif project_ids is not None:
            if len(project_ids) > 1:
                raise ValueError("Only one project id can be provided.")
            kwargs = get_valid_kwargs(
                kwargs,
                self._predict_projects,
                exclude=["project_ids"],
            )
            self._iterator = self._predict_projects(project_ids, **kwargs)
        else:
            raise ValueError(
                "Unknown input type. Supported types are: numpy array, path to a file or directory, ImageInfo, VideoInfo, ProjectInfo, DatasetInfo."
            )

    def _set_var_from_kwargs(self, key, kwargs, default):
        value = kwargs.get(key, default)
        if value is None:
            return None
        if not isinstance(value, list):
            value = [value]
        return value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        if exc_type is not None:
            return False

    def __next__(self):
        try:
            prediction_json = self._iterator.__next__()
            this_kwargs = next(self.prediction_kwargs_iterator)
            prediction = Prediction.from_json(
                prediction_json, **self.kwargs, **this_kwargs, model_meta=self.model_meta
            )
            return prediction
        except StopIteration:
            self._on_infernce_end()
            raise
        except Exception:
            self.stop()
            raise

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._iterator)

    def _get_api_token(self):
        if self.api is not None:
            return self.api.token
        return env.api_token(raise_not_found=False)
    
    def _get_json_body(self):
        body = {"state": {}, "context": {}}
        if self.inference_request_uuid is not None:
            body["state"]["inference_request_uuid"] = self.inference_request_uuid
        if self.inference_settings:
            body["state"]["settings"] = self.inference_settings
        if self.api_token is not None:
            body["api_token"] = self.api_token
        if "model_prediction_suffix" in self.kwargs:
            body["state"]["model_prediction_suffix"] = self.kwargs["model_prediction_suffix"]
        return body
    
    def _post(self, method, *args, retries=5, **kwargs) -> requests.Response:
        if kwargs.get("headers") is None:
            kwargs["headers"] = {}
        if self.api is not None:
            retries = min(self.api.retry_count, retries)
            if "x-api-key" not in kwargs["headers"]:
                kwargs["headers"]["x-api-key"] = self.api.token
        url = self._base_url.rstrip("/") + "/" + method.lstrip("/")
        if "timeout" not in kwargs:
            kwargs["timeout"] = 60
        for retry_idx in range(retries):
            response = None
            try:
                logger.trace(f"POST {url}")
                response = requests.post(url, *args, **kwargs)
                if response.status_code != requests.codes.ok:  # pylint: disable=no-member
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(
                    logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=5,
                    response=response,
                    retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                )
                if retry_idx + 1 == retries:
                    raise exc

    def _get_session_info(self) -> Dict[str, Any]:
        method = "get_session_info"
        r = self._post(method, json=self._get_json_body())
        return r.json()
    
    def _get_inference_progress(self):
        method = "get_inference_progress"
        r = self._post(method, json=self._get_json_body())
        return r.json()

    def _get_inference_status(self):
        method = "get_inference_status"
        r = self._post(method, json=self._get_json_body())
        return r.json()

    def _stop_async_inference(self):
        method = "stop_inference"
        r = self._post(
            method,
            json=self._get_json_body(),
        )
        logger.info("Inference will be stopped on the server")
        return r.json()

    def _clear_inference_request(self):
        method = "clear_inference_request"
        r = self._post(
            method,
            json=self._get_json_body(),
        )
        logger.info("Inference request will be cleared on the server")
        return r.json()

    def _on_infernce_end(self):
        if self.inference_request_uuid is None:
            return
        self._clear_inference_request()

    @property
    def model_meta(self) -> ProjectMeta:
        if self._model_meta is None:
            self._model_meta = ProjectMeta.from_json(
                self._post("get_model_meta", json=self._get_json_body()).json()
            )
        return self._model_meta

    def stop(self):
        if self.inference_request_uuid is None:
            logger.debug("No active inference request to stop.")
            return
        self._stop_async_inference()
        self._on_infernce_end()

    def is_done(self):
        if self.inference_request_uuid is None:
            raise RuntimeError(
                "Inference is not started. Please start inference before checking the status."
            )
        return not self._get_inference_progress()["is_inferring"]

    def progress(self):
        if self.inference_request_uuid is None:
            raise RuntimeError(
                "Inference is not started. Please start inference before checking the status."
            )
        return self._get_inference_progress()["progress"]

    def status(self):
        if self.inference_request_uuid is None:
            raise RuntimeError(
                "Inference is not started. Please start inference before checking the status."
            )
        return self._get_inference_status()

    def _pop_pending_results(self) -> Dict[str, Any]:
        method = "pop_inference_results"
        json_body = self._get_json_body()
        return self._post(method, json=json_body).json()

    def _update_progress(
        self,
        tqdm: tqdm,
        message: str = None,
        current: int = None,
        total: int = None,
        is_size: bool = False,
    ):
        if tqdm is None:
            return
        refresh = False
        if message is not None and tqdm.desc != message:
            tqdm.set_description(message, refresh=False)
            refresh = True
        if current is not None and tqdm.n != current:
            tqdm.n = current
            refresh = True
        if total is not None and tqdm.total != total:
            tqdm.total = total
            refresh = True
        if is_size and tqdm.unit == "it":
            tqdm.unit = "iB"
            tqdm.unit_scale = True
            tqdm.unit_divisor = 1024
            refresh = True
        if not is_size and tqdm.unit == "iB":
            tqdm.unit = "it"
            tqdm.unit_scale = False
            tqdm.unit_divisor = 1
            refresh = True
        if refresh:
            tqdm.refresh()

    def _update_progress_from_response(self, tqdm: tqdm, response: Dict[str, Any]):
        if tqdm is None:
            return
        json_progress = response.get("progress", None)
        if json_progress is None or json_progress.get("message") is None:
            json_progress = response.get("preparing_progress", None)
        if json_progress is None:
            return
        message = json_progress.get("message", json_progress.get("status", None))
        current = json_progress.get("current", None)
        total = json_progress.get("total", None)
        is_size = json_progress.get("is_size", False)
        self._update_progress(tqdm, message, current, total, is_size)

    def _wait_for_inference_start(
        self, delay=1, timeout=None, tqdm: tqdm = None
    ) -> Tuple[dict, bool]:
        has_started = False
        timeout_exceeded = False
        t0 = time.time()
        last_stage = None
        while not has_started and not timeout_exceeded:
            resp = self._get_inference_progress()
            stage = resp.get("stage")
            if stage != last_stage:
                logger.info(stage)
                last_stage = stage
            has_started = stage not in ["preparing", "preprocess", None]
            has_started = has_started or bool(resp.get("result")) or resp["progress"]["total"] != 1
            self._update_progress_from_response(tqdm, resp)
            if not has_started:
                time.sleep(delay)
            timeout_exceeded = timeout and time.time() - t0 > timeout
        if timeout_exceeded:
            self.stop()
            raise Timeout("Timeout exceeded. The server didn't start the inference")
        return resp, has_started

    def _wait_for_pending_results(self, delay=1, timeout=600, tqdm: tqdm = None) -> List[dict]:
        logger.debug("waiting pending results...")
        has_results = False
        timeout_exceeded = False
        t0 = time.monotonic()
        while not has_results and not timeout_exceeded:
            resp = self._pop_pending_results()
            self._update_progress_from_response(tqdm, resp)
            pending_results = resp["pending_results"]
            exception_json = resp["exception"]
            if exception_json:
                exception_str = f"{exception_json['type']}: {exception_json['message']}"
                raise RuntimeError(f"Inference Error: {exception_str}")
            has_results = bool(pending_results)
            if resp.get("finished", False):
                break
            if not has_results:
                time.sleep(delay)
            timeout_exceeded = timeout and time.monotonic() - t0 > timeout
        if timeout_exceeded:
            self.stop()
            raise Timeout("Timeout exceeded. Pending results not received from the server.")
        return pending_results

    def _start_inference(self, method, **kwargs):
        if self.inference_request_uuid:
            raise RuntimeError(
                "Inference is already running. Please stop it before starting a new one."
            )
        resp = self._post(method, **kwargs).json()

        self.inference_request_uuid = resp["inference_request_uuid"]

        logger.info(
            "Inference has started:",
            extra={"inference_request_uuid": resp.get("inference_request_uuid")},
        )
        try:
            resp, has_started = self._wait_for_inference_start(tqdm=self.tqdm)
        except:
            self.stop()
            raise
        frame_iterator = self.Iterator(resp["progress"]["total"], self, tqdm=self.tqdm)
        return frame_iterator

    def _predict_images(self, images: List, **kwargs: dict):
        if isinstance(images[0], bytes):
            f = self._predict_images_bytes
        elif isinstance(images[0], (str, PathLike)):
            f = self._predict_images_paths
        elif isinstance(images[0], np.ndarray):
            f = self._predict_images_nps
        elif isinstance(images[0], int):
            f = self._predict_images_ids
        else:
            raise ValueError(f"Unsupported input type '{type(images[0])}'.")
        kwargs = get_valid_kwargs(kwargs, f, exclude=["images"])
        return f(images, **kwargs)

    def _predict_images_bytes(self, images: List[bytes], batch_size: int = None):
        files = [
            ("files", (f"image_{i}.png", image, "image/png")) for i, image in enumerate(images)
        ]
        state = self._get_json_body()["state"]
        if batch_size is not None:
            state["batch_size"] = batch_size
        method = "inference_batch_async"
        uploads = files + [("state", (None, json.dumps(state), "text/plain"))]
        return self._start_inference(method, files=uploads)

    def _predict_images_paths(self, images: List, batch_size: int = None):
        files = []
        try:
            files = [("files", open(f, "rb")) for f in images]
            state = self._get_json_body()["state"]
            if batch_size is not None:
                state["batch_size"] = batch_size
            method = "inference_batch_async"
            uploads = files + [("state", (None, json.dumps(state), "text/plain"))]
            return self._start_inference(method, files=uploads)
        finally:
            for _, f in files:
                f.close()

    def _predict_images_nps(self, images: List, batch_size: int = None):
        images = [write_bytes(image, ".png") for image in images]
        return self._predict_images_bytes(images, batch_size=batch_size)

    def _predict_images_ids(
        self,
        images: List[int],
        batch_size: int = None,
        upload_mode: str = None,
        output_project_id: int = None,
    ):
        method = "inference_batch_ids_async"
        json_body = self._get_json_body()
        state = json_body["state"]
        state["images_ids"] = images
        if batch_size is not None:
            state["batch_size"] = batch_size
        if upload_mode is not None:
            state["upload_mode"] = upload_mode
        if output_project_id is not None:
            state["output_project_id"] = output_project_id
        return self._start_inference(method, json=json_body)

    def _predict_videos(
        self,
        videos: Union[List[int], List[str], List[PathLike]],
        start_frame: int = None,
        num_frames: int = None,
        stride=None,
        end_frame=None,
        duration=None,
        direction: Literal["forward", "backward"] = None,
        tracker: Literal["botsort"] = None,
        tracker_settings: dict = None,
        batch_size: int = None,
    ):
        if len(videos) != 1:
            raise ValueError("Only one video can be processed at a time.")
        json_body = self._get_json_body()
        state = json_body["state"]
        for key, value in (
            ("start_frame", start_frame),
            ("num_frames", num_frames),
            ("stride", stride),
            ("end_frame", end_frame),
            ("duration", duration),
            ("direction", direction),
            ("tracker", tracker),
            ("tracker_settings", tracker_settings), 
            ("batch_size", batch_size),
        ):
            if value is not None:
                state[key] = value
        if isinstance(videos[0], int):
            method = "inference_video_id_async"
            state["video_id"] = videos[0]
            return self._start_inference(method, json=json_body)
        elif isinstance(videos[0], (str, PathLike)):
            video_path = videos[0]
            files = []
            try:
                method = "inference_video_async"
                files.append((Path(video_path).name, open(video_path, "rb"), "video/*"))
                fields = {
                    "files": files[-1],
                    "state": json.dumps(state),
                }
                encoder = MultipartEncoder(fields)
                if self.tqdm is not None:

                    def _callback(monitor):
                        self.tqdm.update(monitor.bytes_read)

                    video_size = get_file_size(video_path)
                    self._update_progress(self.tqdm, "Uploading video", 0, video_size, is_size=True)
                    encoder = MultipartEncoderMonitor(encoder, _callback)

                return self._start_inference(
                    method, data=encoder, headers={"Content-Type": encoder.content_type}
                )
            finally:
                for _, f, _ in files:
                    f.close()
        else:
            raise ValueError(
                f"Unsupported input type '{type(videos[0])}'. Supported types are: int, str, PathLike."
            )

    def _predict_projects(
        self,
        project_ids: List[int],
        dataset_ids: List[int] = None,
        batch_size: int = None,
        upload_mode: str = None,
        iou_merge_threshold: float = None,
        cache_project_on_model: bool = None,
        output_project_id: int = None,
    ):
        if len(project_ids) != 1:
            raise ValueError("Only one project can be processed at a time.")
        method = "inference_project_id_async"
        json_body = self._get_json_body()
        state = json_body["state"]
        state["project_id"] = project_ids[0]
        if dataset_ids is not None:
            state["dataset_ids"] = dataset_ids
        if batch_size is not None:
            state["batch_size"] = batch_size
        if upload_mode is not None:
            state["upload_mode"] = upload_mode
        if iou_merge_threshold is not None:
            state["iou_merge_threshold"] = iou_merge_threshold
        if cache_project_on_model is not None:
            state["cache_project_on_model"] = cache_project_on_model
        if output_project_id is not None:
            state["output_project_id"] = output_project_id
        return self._start_inference(method, json=json_body)

    def _predict_datasets(
        self,
        dataset_ids: List[int],
        batch_size: int = None,
        upload_mode: str = None,
        iou_merge_threshold: float = None,
        cache_datasets_on_model: bool = None,
    ):
        if self.api is None:
            raise ValueError("Api is required to use this method.")
        dataset_infos = [self.api.dataset.get_info_by_id(dataset_id) for dataset_id in dataset_ids]
        if len(set([info.project_id for info in dataset_infos])) > 1:
            raise ValueError("All datasets must belong to the same project.")
        return self._predict_projects(
            [dataset_infos[0].project_id],
            dataset_ids=dataset_ids,
            batch_size=batch_size,
            upload_mode=upload_mode,
            iou_merge_threshold=iou_merge_threshold,
            cache_project_on_model=cache_datasets_on_model,
        )



# coding: utf-8
from copy import deepcopy

from supervisely.nn.legacy.hosted.constants import MODEL

# Copied from supervisely.nn.hosted.inference_modes
# TODO factor out to a separate common set of constants
MODEL_CLASSES = 'model_classes'
MODEL_SETTINGS = 'model_settings'
NAME = 'name'

MODE = 'mode'
SOURCE = 'source'


def maybe_convert_from_v1_inference_task_config(raw_config):
    if MODEL in raw_config:
        # If the config has sections with new (not v1) names, do not try to convert - most likely the config is
        # already in the new format.
        return raw_config
    else:
        config = {MODE: deepcopy(raw_config.get(MODE, {}))}
        if SOURCE in config[MODE]:
            config[MODE][NAME] = config[MODE].pop(SOURCE)
        if MODEL_CLASSES not in config[MODE]:
            config[MODE][MODEL_CLASSES] = deepcopy(raw_config.get(MODEL_CLASSES, {}))
        config[MODEL] = {}
        for k, v in raw_config.items():
            if k not in (MODE, MODEL_CLASSES):
                config[MODEL][k] = deepcopy(v)
        return config


def maybe_convert_from_deploy_task_config(raw_config):
    if 'connection' in raw_config and MODEL_SETTINGS in raw_config:
        raw_config[MODEL] = raw_config[MODEL_SETTINGS]
    return raw_config




import json
import os
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from supervisely import batched
from supervisely._utils import abs_url, is_development, logger
from supervisely.api.api import Api, ApiField
from supervisely.api.project_api import ProjectInfo
from supervisely.app.exceptions import show_dialog
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.dropdown_checkbox_selector.dropdown_checkbox_selector import (
    DropdownCheckboxSelector,
)
from supervisely.app.widgets.fast_table.fast_table import FastTable
from supervisely.app.widgets.flexbox.flexbox import Flexbox
from supervisely.app.widgets.project_thumbnail.project_thumbnail import ProjectThumbnail
from supervisely.app.widgets.select.select import Select
from supervisely.app.widgets.text.text import Text
from supervisely.app.widgets.widget import Widget
from supervisely.io import env
from supervisely.io.fs import get_file_name_with_ext
from supervisely.nn.experiments import ExperimentInfo


class ExperimentSelector(Widget):
    """
    Widget for selecting experiments from a team.
    """

    class COLUMN:
        NAME = "TASK ID"
        MODEL = "MODEL"
        TRAINING_DATA = "TRAINING DATA"
        CHECKPOINTS = "CHECKPOINTS"
        SESSION = "SESSION"
        BENCHMARK = "BENCHMARK"

    COLUMNS = [
        COLUMN.NAME,
        COLUMN.MODEL,
        COLUMN.TRAINING_DATA,
        COLUMN.CHECKPOINTS,
        COLUMN.SESSION,
        COLUMN.BENCHMARK,
    ]

    class ModelRow:
        def __init__(
            self,
            api: Api,
            team_id: int,
            task_type: str,
            experiment_info: ExperimentInfo,
            project_info: Optional[ProjectInfo] = None,
        ):
            self._api = api
            self._team_id = team_id
            self._task_type = task_type
            self._experiment_info = experiment_info
            self._project_info = project_info

            task_id = experiment_info.task_id
            if task_id == -1:
                pass
            elif task_id == "debug-session":
                task_id = -1
            elif type(task_id) is str:
                if task_id.isdigit():
                    task_id = int(task_id)
                else:
                    raise ValueError(f"Task id {task_id} is not a number")

            # col 1 task
            self._task_id = task_id
            self._task_path = experiment_info.artifacts_dir
            self._task_date = experiment_info.datetime
            self._task_link = self._create_task_link()
            self._config_path = experiment_info.model_files.get("config")
            if self._config_path is not None:
                self._config_path = os.path.join(experiment_info.artifacts_dir, self._config_path)

            # col 2 model
            self._model_name = experiment_info.model_name

            # col 3 project
            self._training_project_id = experiment_info.project_id
            if self._training_project_id is None:
                self._training_project_info = None
            else:
                self._training_project_info = self._project_info

            # col 4 checkpoints
            self._checkpoints = experiment_info.checkpoints

            self._checkpoints_names = []
            self._checkpoints_paths = []
            self._best_checkpoint_value = None
            for checkpoint_path in self._checkpoints:
                self._checkpoints_names.append(get_file_name_with_ext(checkpoint_path))
                self._checkpoints_paths.append(
                    os.path.join(experiment_info.artifacts_dir, checkpoint_path)
                )
                if experiment_info.best_checkpoint == get_file_name_with_ext(checkpoint_path):
                    self._best_checkpoint = os.path.join(
                        experiment_info.artifacts_dir, checkpoint_path
                    )

            # col 5 session
            self._session_link = self._generate_session_link()

            # col 6 benchmark report
            self._benchmark_report_id = experiment_info.evaluation_report_id

            # widgets
            self._task_widget = self._create_task_widget()
            self._model_wiget = self._create_model_widget()
            self._training_project_widget = self._create_training_project_widget()
            self._checkpoints_widget = self._create_checkpoints_widget()
            self._session_widget = self._create_session_widget()
            self._benchmark_widget = self._create_benchmark_widget()

        @property
        def task_id(self) -> int:
            return self._task_id

        @property
        def task_date(self) -> str:
            return self._task_date

        @property
        def task_link(self) -> str:
            return self._task_link

        @property
        def task_type(self) -> str:
            return self._task_type

        @property
        def training_project_info(self) -> ProjectInfo:
            return self._training_project_info

        @property
        def checkpoints_names(self) -> List[str]:
            return self._checkpoints_names

        @property
        def checkpoints_paths(self) -> List[str]:
            return self._checkpoints_paths

        @property
        def checkpoints_selector(self) -> Select:
            return self._checkpoints_widget

        @property
        def session_link(self) -> str:
            return self._session_link

        @property
        def config_path(self) -> str:
            return self._config_path

        def get_selected_checkpoint_path(self) -> str:
            return self._checkpoints_widget.get_value()

        def get_selected_checkpoint_name(self) -> str:
            return self._checkpoints_widget.get_label()

        def set_selected_checkpoint_by_name(self, checkpoint_name: str):
            for i, name in enumerate(self._checkpoints_names):
                if name == checkpoint_name:
                    self._checkpoints_widget.set_value(self._checkpoints_paths[i])
                    return

        def set_selected_checkpoint_by_path(self, checkpoint_path: str):
            for i, path in enumerate(self._checkpoints_paths):
                if path == checkpoint_path:
                    self._checkpoints_widget.set_value(path)
                    return

        def to_html(self) -> List[str]:
            return [
                f"<div> {self._task_widget.to_html()} </div>",
                f"<div> {self._model_wiget.to_html()} </div>",
                f"<div> {self._training_project_widget.to_html()} </div>",
                f"<div> {self._checkpoints_widget.to_html()} </div>",
                f"<div> {self._session_widget.to_html()} </div>",
                f"<div> {self._benchmark_widget.to_html()} </div>",
            ]

        def _create_task_link(self) -> str:
            remote_path = os.path.join(self._task_path, "open_app.lnk")
            task_file = self._api.file.get_info_by_path(self._team_id, remote_path)
            if task_file is not None:
                if is_development():
                    return abs_url(f"/files/{task_file.id}")
                else:
                    return f"/files/{task_file.id}"
            else:
                return ""

        def _generate_session_link(self) -> str:
            if is_development():
                session_link = abs_url(f"/apps/sessions/{self._task_id}")
            else:
                session_link = f"/apps/sessions/{self._task_id}"
            return session_link

        def _create_task_widget(self) -> Flexbox:
            task_widget = Container(
                [
                    Text(
                        f"<i class='zmdi zmdi-folder' style='color: #7f858e'></i> <a href='{self._task_link}' target='_blank'>{self._task_id}</a>",
                        "text",
                    ),
                    Text(
                        f"<span class='field-description text-muted' style='color: #7f858e'>{self._task_date}</span>",
                        "text",
                        font_size=13,
                    ),
                ],
                gap=0,
            )
            return task_widget

        def _create_model_widget(self) -> Text:
            if self._model_name is None:
                self._model_name = "Unknown model"

            model_widget = Text(
                f"<span class='field-description text-muted' style='color: #7f858e'>{self._model_name}</span>",
                "text",
                font_size=13,
            )
            return model_widget

        def _create_training_project_widget(self) -> Union[ProjectThumbnail, Text]:
            training_project_thumbnail = ProjectThumbnail(
                self._training_project_info, remove_margins=True
            )
            training_project_text = Text(
                f"<span class='field-description text-muted' style='color: #7f858e'>Project was deleted</span>",
                "text",
                font_size=13,
            )
            if self.training_project_info is not None:
                training_project_thumbnail.show()
                training_project_text.hide()
            else:
                training_project_thumbnail.hide()
                training_project_text.show()
            return Container(widgets=[training_project_thumbnail, training_project_text], gap=0)

        def checkpoint_changed(self, checkpoint_value: str):
            return

        def _create_checkpoints_widget(self) -> Select:
            checkpoint_selector_items = []
            for path, name in zip(self._checkpoints_paths, self._checkpoints_names):
                checkpoint_selector_items.append(Select.Item(value=path, label=name))
            checkpoint_selector = Select(items=checkpoint_selector_items)
            if self._best_checkpoint_value is not None:
                checkpoint_selector.set_value(self._best_checkpoint)

            @checkpoint_selector.value_changed
            def on_checkpoint_changed(checkpoint_value: str):
                self.checkpoint_changed(checkpoint_value)

            return checkpoint_selector

        def _create_session_widget(self) -> Text:
            session_link_widget = Text(
                f"<a href='{self._session_link}' target='_blank'>Preview</a> <i class='zmdi zmdi-open-in-new'></i>",
                "text",
            )
            return session_link_widget

        def _create_benchmark_widget(self) -> Text:
            if self._benchmark_report_id is None:
                self._benchmark_report_id = "No evaluation report available"
                benchmark_widget = Text(
                    "<span class='field-description text-muted' style='color: #7f858e'>No evaluation report available</span>",
                    "text",
                    font_size=13,
                )
            else:
                if is_development():
                    benchmark_report_link = abs_url(
                        f"/model-benchmark?id={self._benchmark_report_id}"
                    )
                else:
                    benchmark_report_link = f"/model-benchmark?id={self._benchmark_report_id}"

                benchmark_widget = Text(
                    f"<i class='zmdi zmdi-chart' style='color: #7f858e'></i> <a href='{benchmark_report_link}' target='_blank'>Evaluation report</a>",
                    "text",
                )
            return benchmark_widget

        def _widget_to_cell_value(self, widget: Widget) -> str:
            if isinstance(widget, Container):
                return json.dumps(
                    {
                        "widget_id": widget.widget_id,
                        "widgets": [w.widget_id for w in widget._widgets],
                    }
                )
            else:
                return json.dumps({"widget_id": widget.widget_id, "widgets": []})

        def to_table_row(self):
            return [
                self._widget_to_cell_value(w)
                for w in [
                    self._task_widget,
                    self._model_wiget,
                    self._training_project_widget,
                    self._checkpoints_widget,
                    self._session_widget,
                    self._benchmark_widget,
                ]
            ]

        @classmethod
        def widgets_templates(cls):
            checkpoints_template_widget = Select(items=[])
            checkpoints_template_widget.value_changed(lambda _: None)

            return [
                # _task_widget
                Container(widgets=[Text(""), Text("")], gap=0),
                # _model_wiget
                Text(""),
                # _training_project_widget
                Container(widgets=[ProjectThumbnail(remove_margins=True), Text("")], gap=0),
                # _checkpoints_widget
                checkpoints_template_widget,
                # _session_widget
                Text(""),
                # _benchmark_widget
                Text(""),
            ]

        def search_text(self) -> str:
            text = ""
            text += str(self._task_id)
            text += str(self._task_date)
            text += str(self._model_name)
            if self._training_project_info is not None:
                text += str(self._training_project_info.name)
            else:
                text += "Project was deleted"
            return text

        def sort_values(self) -> List[int]:
            # Sort by training project name: real names first (A->Z), deleted projects last
            if self._training_project_info is not None:
                training_project_name = (0, self._training_project_info.name.lower())
            else:
                training_project_name = (1, "")

            if self._benchmark_report_id == "No evaluation report available":
                benchmark_report_id = 0
            else:
                benchmark_report_id = 1

            return [
                self._task_id,
                self._model_name.capitalize(),
                training_project_name,
                0,
                0,
                benchmark_report_id,
            ]

    def __init__(
        self,
        api: Api = None,
        team_id: int = None,
        experiment_infos: List[ExperimentInfo] = [],
        widget_id: str = None,
    ):
        if team_id is None:
            team_id = env.team_id()
        self.team_id = team_id
        if api is None:
            api = Api()
        self.api = api
        self._experiment_infos = experiment_infos
        self._checkpoint_changed_func = None

        self._rows = []
        self.table = self._create_table()
        self._rows_search_texts = []
        self._rows_sort_values = []

        self._project_infos_map = self._get_project_infos_map(experiment_infos)
        self.set_experiment_infos(experiment_infos)
        super().__init__(widget_id=widget_id)

    def _search_function(self, data: pd.DataFrame, search_value: str) -> List[ModelRow]:
        search_texts = []
        for idx in data.index:
            first_col_value = data.loc[idx, self.COLUMNS[0]]
            if isinstance(first_col_value, pd.Series):
                first_col_value = first_col_value.iloc[0]
            original_idx = self._first_column_value_to_index[first_col_value]
            search_texts.append(self._rows_search_texts[original_idx])

        search_series = pd.Series(search_texts, index=data.index)
        mask = search_series.str.contains(search_value, case=False, na=False)
        return data[mask]

    def _sort_function(
        self, data: pd.DataFrame, column_idx: int, order: str = "asc"
    ) -> List[ModelRow]:
        data = data.copy()
        if column_idx >= len(self._rows_sort_values[0]) if self._rows_sort_values else True:
            raise IndexError(
                f"Sorting by column idx = {column_idx} is not possible, your sort values have only {len(self._rows_sort_values[0]) if self._rows_sort_values else 0} columns with idx from 0 to {len(self._rows_sort_values[0]) - 1 if self._rows_sort_values else -1}"
            )

        if order == "asc":
            ascending = True
        else:
            ascending = False

        try:
            sort_values = []
            for idx in data.index:
                first_col_value = data.loc[idx, self.COLUMNS[0]]
                if isinstance(first_col_value, pd.Series):
                    first_col_value = first_col_value.iloc[0]
                original_idx = self._first_column_value_to_index[first_col_value]
                sort_values.append(self._rows_sort_values[original_idx][column_idx])

            sort_series = pd.Series(sort_values, index=data.index)
            sorted_indices = sort_series.sort_values(ascending=ascending).index
            data = data.loc[sorted_indices]
            data.reset_index(inplace=True, drop=True)

        except IndexError as e:
            e.args = (
                f"Sorting by column idx = {column_idx} is not possible, your sort values have only {len(self._rows_sort_values[0]) if self._rows_sort_values else 0} columns with idx from 0 to {len(self._rows_sort_values[0]) - 1 if self._rows_sort_values else -1}",
            )
            raise e

        return data

    def _filter_function(
        self, data: pd.DataFrame, filter_value: Tuple[List[str], List[str]]
    ) -> pd.DataFrame:
        try:
            frameworks, task_types = filter_value

            filtered_experiments_idxs = set()
            if not frameworks and not task_types:
                return data

            for idx, experiment_info in enumerate(self._experiment_infos):
                should_add = True
                if frameworks and experiment_info.framework_name not in frameworks:
                    should_add = False
                if task_types and experiment_info.task_type not in task_types:
                    should_add = False
                if should_add:
                    filtered_experiments_idxs.add(idx)

            filtered_data = data.iloc[sorted(filtered_experiments_idxs)]
            filtered_data.reset_index(inplace=True, drop=True)
            return filtered_data
        except Exception as e:
            logger.error(f"Error during filtering: {e}", exc_info=True)
            show_dialog(title="Filtering Error", description=str(e), status="error")
            return data

    def _get_frameworks(self):
        frameworks = set()
        for experiment_info in self._experiment_infos:
            frameworks.add(experiment_info.framework_name)
        return sorted(frameworks)

    def _get_task_types(self):
        task_types = set()
        for experiment_info in self._experiment_infos:
            task_types.add(experiment_info.task_type)
        return sorted(task_types)

    def _create_table(self) -> FastTable:
        widgets = self.ModelRow.widgets_templates()
        columns = []
        columns_options = []
        for column_name, widget in zip(self.COLUMNS, widgets):
            columns.append((column_name, widget))
            columns_options.append({"customCell": True})
        columns_options[3].update({"classes": "border border-gray-200 px-2"})
        columns_options[3].update({"disableSort": True})
        columns_options[4].update({"disableSort": True})
        self.framework_filter = DropdownCheckboxSelector(
            label="Framework",
            items=[
                DropdownCheckboxSelector.Item(framework) for framework in self._get_frameworks()
            ],
        )
        self.task_type_filter = DropdownCheckboxSelector(
            label="Task Type",
            items=[
                DropdownCheckboxSelector.Item(task_type) for task_type in self._get_task_types()
            ],
        )
        table = FastTable(
            columns=columns,
            columns_options=columns_options,
            is_radio=True,
            page_size=10,
            header_right_content=Container(
                widgets=[self.framework_filter, self.task_type_filter],
                gap=10,
                direction="horizontal",
            ),
        )
        table.set_search(self._search_function)
        table.set_sort(self._sort_function)
        table.set_filter(self._filter_function)

        @self.framework_filter.value_changed
        def on_framework_filter_change(
            selected_frameworks: List[DropdownCheckboxSelector.Item],
        ):
            selected_frameworks = [item.id for item in selected_frameworks]
            selected_task_types = self.task_type_filter.get_selected()
            self.table.filter((selected_frameworks, selected_task_types))

        @self.task_type_filter.value_changed
        def on_task_type_filter_change(
            selected_task_types: List[DropdownCheckboxSelector.Item],
        ):
            selected_task_types = [item.id for item in selected_task_types]
            selected_frameworks = self.framework_filter.get_selected()
            self.table.filter((selected_frameworks, selected_task_types))

        return table

    def _get_project_infos_map(
        self, experiment_infos: List[ExperimentInfo]
    ) -> Dict[int, ProjectInfo]:
        """
        Returns a map of project IDs to project infos used in the experiment infos.
        """
        project_ids = set()
        for experiment_info in experiment_infos:
            if experiment_info.project_id is not None:
                project_ids.add(experiment_info.project_id)
        project_ids = list(project_ids)

        project_infos_map = {}
        if project_ids is not None:
            for batch in batched(project_ids):
                filters = [
                    {
                        ApiField.FIELD: ApiField.ID,
                        ApiField.OPERATOR: "in",
                        ApiField.VALUE: batch,
                    },
                ]

                fields = [ApiField.IMAGES_COUNT, ApiField.REFERENCE_IMAGE_URL]
                batch_infos = self.api.project.get_list(
                    team_id=self.team_id,
                    filters=filters,
                    fields=fields,
                )
                for info in batch_infos:
                    project_infos_map[info.id] = info

        return project_infos_map

    def _generate_table_rows(self, experiment_infos: List[ExperimentInfo]) -> List[ModelRow]:
        """Method to generate table rows from remote path to training app save directory"""

        def process_experiment_info(experiment_info: ExperimentInfo):
            try:
                logger.debug(f"Processing experiment info: {experiment_info.task_id}")
                project_info = self._project_infos_map.get(experiment_info.project_id)
                model_row = ExperimentSelector.ModelRow(
                    api=self.api,
                    team_id=self.team_id,
                    task_type=experiment_info.task_type,
                    experiment_info=experiment_info,
                    project_info=project_info,
                )

                def this_row_checkpoint_changed(checkpoint_value: str):
                    self._checkpoint_changed(model_row, checkpoint_value)

                model_row.checkpoint_changed = this_row_checkpoint_changed
                return experiment_info.task_type, model_row
            except Exception as e:
                logger.debug(f"Failed to process experiment info: {experiment_info}")
                return None, None

        table_rows = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [
                executor.submit(process_experiment_info, experiment_info)
                for experiment_info in experiment_infos
            ]

            for future in futures:
                result = future.result()
                if result:
                    task_type, model_row = result
                    if task_type is not None and model_row is not None:
                        table_rows.append(model_row)

        table_rows.sort(key=lambda x: x.task_id, reverse=True)
        return table_rows

    def _update_search_text(self):
        self._rows_search_texts = [row.search_text() for row in self._rows]

    def _update_sort_values(self):
        self._rows_sort_values = [row.sort_values() for row in self._rows]

    def _update_value_index_map(self):
        self._first_column_value_to_index = {}
        for i, row in self.table._source_data.iterrows():
            value = row.iloc[0]
            self._first_column_value_to_index[value] = i

    def set_experiment_infos(self, experiment_infos: List[ExperimentInfo]) -> None:
        """
        Updates the experiment infos and regenerates the table rows.
        """
        table_rows = self._generate_table_rows(experiment_infos)
        self._rows = table_rows
        for row in table_rows:
            self.table.insert_row(row.to_table_row())
        self._update_value_index_map()
        self._update_search_text()
        self._update_sort_values()

    def get_selected_experiment_info(self) -> Union[ExperimentInfo, None]:
        selected_row = self.table.get_selected_row()
        if selected_row is None:
            return None
        return self._rows[selected_row.row_index]._experiment_info

    def get_selected_experiment_info_json(self) -> Union[dict, None]:
        experiment_info = self.get_selected_experiment_info()
        if experiment_info is None:
            return None
        return experiment_info.to_json()

    def get_selected_checkpoint_name(self) -> Union[str, None]:
        selected_row = self.table.get_selected_row()
        if selected_row is None:
            return None
        return self._rows[selected_row.row_index].get_selected_checkpoint_name()

    def get_selected_checkpoint_path(self) -> Union[str, None]:
        selected_row = self.table.get_selected_row()
        if selected_row is None:
            return None
        return self._rows[selected_row.row_index].get_selected_checkpoint_path()

    def set_selected_row_by_experiment_info(self, experiment_info: ExperimentInfo) -> None:
        for idx, row in enumerate(self._rows):
            if row._experiment_info.task_id == experiment_info.task_id:
                self.table.select_row(idx)
                return
        raise ValueError(f"Experiment info {experiment_info} not found in the table rows.")

    def _checkpoint_changed(self, row: ModelRow, checkpoint_value: str):
        if self._checkpoint_changed_func is None:
            return
        return self._checkpoint_changed_func(row, checkpoint_value)

    def checkpoint_changed(self, func: Callable[[ModelRow, str], None]):
        self._checkpoint_changed_func = func
        return self._checkpoint_changed_func

    def selection_changed(self, func):
        def f(selected_row: FastTable.ClickedRow):
            if selected_row is None:
                return
            idx = selected_row.row_index
            experiment_info = self._rows[idx]._experiment_info
            func(experiment_info)

        return self.table.selection_changed(f)

    def set_selected_checkpoint_by_name(self, checkpoint_name: str):
        selected_row = self.table.get_selected_row()
        if selected_row is None:
            return
        self._rows[selected_row.row_index].set_selected_checkpoint_by_name(checkpoint_name)

    def set_selected_row_by_task_id(self, task_id: int):
        for idx, row in enumerate(self._rows):
            if row._experiment_info.task_id == task_id:
                self.table.select_row(idx)
                return
        raise ValueError(f"Experiment info with task id {task_id} not found in the table rows.")

    def get_selected_row_by_task_id(self, task_id: int):
        for idx, row in enumerate(self._rows):
            if row._experiment_info.task_id == task_id:
                return row
        return None

    def search(self, search_value: str):
        self.table.search(search_value)

    def disable(self):
        return self.table.disable()

    def enable(self):
        return self.table.enable()

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}

    def to_html(self):
        return self.table.to_html()



from typing import Callable, Dict, List, Optional, Union

import supervisely.io.env as env
from supervisely.api.api import Api
from supervisely.app.widgets import Widget
from supervisely.app.widgets.checkbox.checkbox import Checkbox
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.select.select import Select
from supervisely.app.widgets.tree_select.tree_select import TreeSelect
from supervisely.project.project_type import ProjectType


class SelectCollection(Widget):
    """SelectCollection widget in Supervisely is a widget that allows users to select one or multiple collections (entities collections).
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/selection/selectCollection>`_
        (including screenshots and examples).

    :param default_id: The ID of the collection to be selected by default.
    :type default_id: Union[int, None]
    :param project_id: The ID of the project to read collections from.
    :type project_id: Union[int, None]
    :param multiselect: Whether multiple collections can be selected.
    :type multiselect: bool
    :param compact: Whether the widget should be compact (e.g. no team, workspace, and project selectors).
    :type compact: bool
    :param select_all_collections: Whether all collections should be selected by default.
    :type select_all_collections: bool
    :param allowed_project_types: The list of project types that are allowed to be selected.
    :type allowed_project_types: Optional[List[ProjectType]]
    :param team_is_selectable: Whether the team selector should be selectable.
    :type team_is_selectable: bool
    :param workspace_is_selectable: Whether the workspace selector should be selectable.
    :type workspace_is_selectable: bool
    :param widget_id: The unique identifier of the widget.
    :type widget_id: Union[str, None]
    :param show_select_all_collections_checkbox: Whether the checkbox to select all collections should be shown.
    :type show_select_all_collections_checkbox: bool

    :Public methods:
    - `set_project_id(project_id: int) -> None`: Set the project ID to read collections from.
    - `get_selected_ids() -> Optional[List[int]]`: Get the IDs of the selected collections.
    - `get_selected_id() -> Optional[int]`: Get the ID of the selected collection.
    - `value_changed(func: Callable) -> Callable`: Decorator to set the callback function for the value changed event.
    - `set_collection_id(collection_id: int) -> None`: Set the ID of the collection to be selected by default.
    - `set_collection_ids(collection_ids: List[int]) -> None`: Set the IDs of the collections to be selected by default.
    - `get_selected_project_id() -> Optional[int]`: Get the ID of the selected project.
    - `get_selected_team_id() -> int`: Get the ID of the selected team.
    - `set_team_id(team_id: int) -> None`: Set the team ID to read workspaces from.
    - `get_selected_workspace_id() -> int`: Get the ID of the selected workspace.
    - `set_workspace_id(workspace_id: int) -> None`: Set the workspace ID to read projects from.
    - `is_all_selected() -> bool`: Check if all collections are selected.
    - `select_all() -> None`: Select all collections.
    - `disable() -> None`: Disable the widget in the UI.
    - `enable() -> None`: Enable the widget in the UI.
    - `set_selected_id(collection_id: int) -> None`: Set the ID of the collection to be selected by default.
    - `set_selected_ids(collection_ids: List[int]) -> None`: Set the IDs of the collections to be selected by default.

    :Properties:
    - `team_id`: The ID of the team selected in the widget.
    - `workspace_id`: The ID of the workspace selected in the widget.
    - `project_id`: The ID of the project selected in the widget.

    :Usage example:

        .. code-block:: python
            from supervisely.app.widgets import SelectCollection

            project_id = 123
            collection_id = 456

            select_collection = SelectCollection(
                default_id=collection_id,
                project_id=project_id,
                multiselect=True,
            )

            @select_collection.value_changed
            def on_change(selected_ids):
                print(selected_ids) # Output: [456, 789]
    """

    def __init__(
        self,
        default_id: Union[int, None] = None,
        project_id: Union[int, None] = None,
        multiselect: bool = False,
        compact: bool = False,
        select_all_collections: bool = False,
        allowed_project_types: Optional[List[ProjectType]] = None,
        team_is_selectable: bool = True,
        workspace_is_selectable: bool = True,
        show_select_all_collections_checkbox: bool = True,
        widget_id: Union[str, None] = None,
        width: int = 193,
    ):
        self._api = Api.from_env()

        if default_id is not None and project_id is None:
            raise ValueError("Project ID must be provided when default collection ID is set.")

        if not multiselect and select_all_collections:
            raise ValueError("Select all collections is only available in multiselect mode.")

        # Reading team_id and workspace_id from environment variables.
        # If not found, error will be raised.
        self._team_id = env.team_id()
        self._workspace_id = env.workspace_id()

        # Using environment variables to set the default values if they are not provided.
        self._project_id = project_id or env.project_id(raise_not_found=False)
        self._collection_id = default_id

        # Get mapping of collection ID to name for current project.
        self._collections_names_map = None
        self._collections_ids_map = None

        self._multiselect = multiselect
        self._compact = compact

        # Extract values from Enum to match the .type property of the ProjectInfo object.

        self._project_types = None
        if allowed_project_types is not None:
            if all(allowed_project_types) is isinstance(allowed_project_types, ProjectType):
                self._project_types = (
                    [project_type.value for project_type in allowed_project_types]
                    if allowed_project_types is not None
                    else None
                )
            elif all(allowed_project_types) is isinstance(allowed_project_types, str):
                self._project_types = allowed_project_types

        # Widget components.
        self._select_team = None
        self._select_workspace = None
        self._select_project = None
        self._select_collection = None
        self._select_all_collections_checkbox = None
        self._width = width

        # List of widgets will be used to create a Container.
        self._widgets = []

        if not compact:
            # If the widget is not compact, create team, workspace, and project selectors.
            self._create_selectors(team_is_selectable, workspace_is_selectable)

        # Create the collection selector.
        self._create_collection_selector(select_all_collections)

        # Create the checkbox to select all collections if needed.
        if show_select_all_collections_checkbox:
            self._create_select_all_collections_checkbox(select_all_collections)

        # Group the selectors and the collection selector into a container.
        self._content = Container(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def disable(self):
        """Disable the widget in the UI."""
        for widget in self._widgets:
            widget.disable()

    def enable(self) -> None:
        """Enable the widget in the UI."""
        for widget in self._widgets:
            widget.enable()

    @property
    def team_id(self) -> int:
        """The ID of the team selected in the widget.

        :return: The ID of the team.
        :rtype: int
        """
        return self._team_id

    @team_id.setter
    def team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        if not self._compact:
            self._select_team.set_value(team_id)
            self._select_workspace.set(self._get_select_items(team_id=team_id))
        self._team_id = team_id

    def get_selected_team_id(self) -> int:
        """Get the ID of the selected team.

        :return: The ID of the selected team.
        :rtype: int
        """
        return self.team_id

    def set_team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        self.team_id = team_id

    @property
    def workspace_id(self) -> int:
        """The ID of the workspace selected in the widget.

        :return: The ID of the workspace.
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        if not self._compact:
            self._select_workspace.set_value(workspace_id)
            self._select_project.set(self._get_select_items(workspace_id=workspace_id))
        self._workspace_id = workspace_id

    def get_selected_workspace_id(self) -> int:
        """Get the ID of the selected workspace.

        :return: The ID of the selected workspace.
        :rtype: int
        """
        return self.workspace_id

    def set_workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        self.workspace_id = workspace_id

    @property
    def project_id(self) -> Optional[int]:
        """The ID of the project selected in the widget.

        :return: The ID of the project.
        :rtype: Optional[int]
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read collections from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        if not self._compact:
            self._select_project.set_value(project_id)
        self.set_project_id(project_id)

    def get_selected_project_id(self) -> Optional[int]:
        """Get the ID of the selected project.

        :return: The ID of the selected project.
        :rtype: Optional[int]
        """
        return self.project_id

    def set_collection(self, collection: Union[int, str]) -> None:
        """Set the collection to be selected.

        :param collection: The ID or name of the collection.
        :type collection: The ID or name of the collection.
        :raise ValueError: If multiselect is enabled.
        """
        if isinstance(collection, int):
            self.set_selected_id(collection)
        elif isinstance(collection, str):
            self.set_selected_name(collection)
        else:
            raise ValueError("Collection ID must be an integer or a string.")

    def set_collections(self, collections: Union[List[int], List[str]]) -> None:
        """Set the collections to be selected.

        :param collection: The ID or name of the collection.
        :type collection: The ID or name of the collection.
        :raise ValueError: If multiselect is disabled.
        """
        if not isinstance(collections, list):
            raise ValueError("Collections must be a list of integers or strings.")
        if all(isinstance(i, int) for i in collections) or len(collections) == 0:
            self.set_selected_ids(collections)
        elif all(isinstance(i, str) for i in collections):
            self.set_selected_names(collections)
        else:
            raise ValueError("Collection IDs must be a list of integers or a list of strings.")

    def value_changed(self, func: Callable) -> Callable:
        """Decorator to set the callback function for the value changed event.

        :param func: The callback function.
        :type func: Callable
        :return: The callback function.
        :rtype: Callable
        """

        @self._select_collection.value_changed
        def _click(items: Union[List[str], str]):
            if isinstance(items, list):
                res = [self._collections_names_map[item].id for item in items]
            else:
                res = self._collections_names_map[items].id

            func(res)

        return _click

    def _create_select_all_collections_checkbox(self, select_all_collections: bool) -> None:
        """Create the checkbox to select all collections.

        :param select_all_collections: Whether all collections should be selected by default.
        :type select_all_collections: bool
        """
        if not self._multiselect:
            # We'll only create the checkbox if multiselect is enabled.
            return
        select_all_collections_checkbox = Checkbox("Select all collections")

        @select_all_collections_checkbox.value_changed
        def select_all_collections_checkbox_handler(checked: bool) -> None:
            """Handler function for the event when the checkbox value changes.

            :param checked: The value of the checkbox.
            :type checked: bool
            """
            if self._project_id is None:
                return

            if checked:
                self.select_all()
            else:
                self.deselect_all()

        self._widgets.append(select_all_collections_checkbox)
        self._select_all_collections_checkbox = select_all_collections_checkbox
        if select_all_collections:
            self.select_all()

    def _create_collection_selector(self, select_all_collections: bool) -> None:
        """Create the collection selector.

        :param select_all_collections: Whether all collections should be selected by default.
        :type select_all_collections: bool
        """
        items = self._read_collections(self._project_id)
        if items is None or len(items) == 0:
            items = []
        self._select_collection = Select(
            items=items,
            multiple=self._multiselect,
            width_px=self._width,
            placeholder="Select collection",
            filterable=True,
        )
        if self._collection_id is not None:
            info = self._collections_ids_map.get(self._collection_id)
            if info is not None:
                value = [info.name] if self._multiselect else info.name
                self._select_collection.set_value(value)
        if select_all_collections:
            self.select_all()

        # Adding the collection selector to the list of widgets to be added to the container.
        self._widgets.append(self._select_collection)

    def _create_selectors(self, team_is_selectable: bool, workspace_is_selectable: bool):
        """Create the team, workspace, and project selectors.

        :param team_is_selectable: Whether the team selector should be selectable.
        :type team_is_selectable: bool
        :param workspace_is_selectable: Whether the workspace selector should be selectable.
        :type workspace_is_selectable: bool
        """

        def team_selector_handler(team_id: int):
            """Handler function for the event when the team selector value changes.

            :param team_id: The ID of the selected team.
            :type team_id: int
            """
            self._select_workspace.set(items=self._get_select_items(team_id=team_id))
            self._team_id = team_id

        def workspace_selector_handler(workspace_id: int):
            """Handler function for the event when the workspace selector value changes.

            :param workspace_id: The ID of the selected workspace.
            :type workspace_id: int
            """
            self._select_project.set(items=self._get_select_items(workspace_id=workspace_id))
            self._workspace_id = workspace_id

        def project_selector_handler(project_id: int):
            """Handler function for the event when the project selector value changes.

            :param project_id: The ID of the selected project.
            :type project_id: int
            """
            self._select_collection.set(self._read_collections(project_id))
            self._project_id = project_id

            if (
                self._select_all_collections_checkbox is not None
                and self._select_all_collections_checkbox.is_checked()
            ):
                self.select_all()
                self._select_collection.hide()

        self._select_team = Select(
            items=self._get_select_items(),
            placeholder="Select team",
            filterable=True,
            width_px=self._width,
        )
        self._select_team.set_value(self._team_id)
        if not team_is_selectable:
            self._select_team.disable()

        self._select_workspace = Select(
            items=self._get_select_items(team_id=self._team_id),
            placeholder="Select workspace",
            filterable=True,
            width_px=self._width,
        )
        self._select_workspace.set_value(self._workspace_id)
        if not workspace_is_selectable:
            self._select_workspace.disable()

        self._select_project = Select(
            items=self._get_select_items(workspace_id=self._workspace_id),
            placeholder="Select project",
            filterable=True,
            width_px=self._width,
        )
        self._select_project.set_value(self._project_id)

        # Register the event handlers.
        self._select_team.value_changed(team_selector_handler)
        self._select_workspace.value_changed(workspace_selector_handler)
        self._select_project.value_changed(project_selector_handler)

        # Adding widgets to the list, so they can be added to the container.
        self._widgets.extend([self._select_team, self._select_workspace, self._select_project])

    def _get_select_items(self, **kwargs) -> List[Select.Item]:
        """Get the list of items for the team, workspace, and project selectors.
        Possible keyword arguments are 'team_id' and 'workspace_id'.

        :return: The list of items.
        :rtype: List[Select.Item]
        """
        if not kwargs:
            items = self._api.team.get_list()
        elif "team_id" in kwargs:
            items = self._api.workspace.get_list(kwargs["team_id"])
        elif "workspace_id" in kwargs:
            projects_list = self._api.project.get_list(kwargs["workspace_id"])
            if self._project_types is not None:
                items = [
                    project for project in projects_list if project.type in self._project_types
                ]  # TODO: Filter project from API, not from here.
            else:
                items = projects_list

        return [Select.Item(value=item.id, label=item.name) for item in items]

    def get_json_data(self) -> Dict:
        """Get the JSON data of the widget.

        :return: The JSON data.
        :rtype: Dict
        """
        return {}

    def get_json_state(self) -> Dict:
        """Get the JSON state of the widget.

        :return: The JSON state.
        :rtype: Dict
        """
        return {}

    def _read_collections(self, project_id: Optional[int]) -> Optional[List[Select.Item]]:
        """Get the lisf of Select.Item objects representing the collection hierarchy.

        :param project_id: The ID of the project.
        :type project_id: Optional[int]
        :return: The list of Select.Item objects.
        :rtype: Optional[List[Select.Item]]
        """
        self._fetch_collections(project_id)

        if not self._collections_names_map or not project_id:
            return None

        collections = [Select.Item(i.name, i.name) for i in self._collections_names_map.values()]
        return collections

    def _fetch_collections(self, project_id: Optional[int]) -> None:
        """Get the mapping of collection name to EntitiesCollectionInfo object.

        :param project_id: The ID of the project.
        :type project_id: Optional[int]
        :return: None
        :rtype: None
        """
        self._collections_names_map = {}
        self._collections_ids_map = {}

        if not project_id:
            return

        for collection in self._api.entities_collection.get_list(project_id):
            self._collections_names_map[collection.name] = collection
            self._collections_ids_map[collection.id] = collection

    def set_project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read collections from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        self._project_id = project_id
        items = self._read_collections(project_id)
        if items is None or len(items) == 0:
            self._select_collection.set([])
            return
        self._select_collection.set(items)
        if self._multiselect:
            if self._select_all_collections_checkbox.is_checked():
                self.select_all()
            else:
                self.deselect_all()
        else:
            self._select_collection.set_value("")

    def _get_selected(self) -> Optional[Union[List[int], int]]:
        """Get the ID of the selected collection(s).

        :return: The ID of the selected collection(s).
        :rtype: Optional[Union[List[int], int]]
        """
        selected = self._select_collection.get_value()
        if not selected:
            return None

        if isinstance(selected, list):
            return [self._collections_names_map[item].id for item in selected]
        else:
            return self._collections_names_map[selected].id

    def get_selected_ids(self) -> Optional[List[int]]:
        """Get the IDs of the selected collections.

        raise ValueError if multiselect is disabled.
        return: The IDs of the selected collections.
        rtype: Optional[List[int]]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        return self._get_selected()

    def get_selected_id(self) -> Optional[int]:
        """Get the ID of the selected collection.

        raise ValueError if multiselect is enabled.
        return: The ID of the selected collection.
        rtype: Optional[int]
        """
        if self._multiselect:
            raise ValueError("This method can only be called when multiselect is disabled.")
        return self._get_selected()

    def set_selected_ids(self, collection_ids: List[int]) -> None:
        """Set the IDs of the collections to be selected by default.

        :param collection_ids: The IDs of the collections.
        :type collection_ids: List[int]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._set_selected_by_ids(collection_ids)

    def set_selected_id(self, collection_id: int) -> None:
        """Set the ID of the collection to be selected by default.

        :param collection_id: The ID of the collection.
        :type collection_id: int
        """
        if self._multiselect:
            raise ValueError("This method can only be called when multiselect is disabled.")
        self._set_selected_by_ids([collection_id])

    def _set_selected_by_ids(self, collection_ids: List[int]) -> None:
        """Set the IDs of the collections to be selected by default.

        :param collection_ids: The IDs of the collections.
        :type collection_ids: List[int]
        """
        if not self._collections_ids_map:
            return

        selected = []
        for i in collection_ids:
            if i in self._collections_ids_map:
                selected.append(self._collections_ids_map[i].name)

        if self._multiselect:
            if len(selected) == len(self._collections_ids_map):
                self.select_all()
            else:
                self.deselect_all()
                self._select_collection.set_value(selected)

        else:
            if len(selected) > 1:
                raise ValueError("More than one collection found, but multiselect is disabled.")
            value = selected[0] if selected else ""
            self._select_collection.set_value(value)

    def set_selected_name(self, collection_name: str) -> None:
        """Set the name of the collection to be selected by default.

        :param collection_name: The name of the collection.
        :type collection_name: str
        """
        if self._multiselect:
            raise ValueError("This method can only be called when multiselect is disabled.")
        self._set_selected_by_names([collection_name])

    def set_selected_names(self, collection_names: List[str]) -> None:
        """Set the names of the collections to be selected by default.

        :param collection_names: The names of the collections.
        :type collection_names: List[str]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._set_selected_by_names(collection_names)

    def _set_selected_by_names(self, collection_names: List[str]) -> None:
        """Set the names of the collections to be selected by default.

        :param collection_names: The names of the collections.
        :type collection_names: List[str]
        """
        if not self._collections_names_map:
            return

        ids = []
        for i in collection_names:
            if i in self._collections_names_map:
                ids.append(self._collections_names_map[i].id)
        self._set_selected_by_ids(ids)

    def is_all_selected(self) -> bool:
        """Check if all collections are selected.

        return: True if all collections are selected, False otherwise.
        rtype: bool
        """
        return len(self._select_collection.get_value()) == len(self._collections_names_map)

    def select_all(self) -> None:
        """Select all collections."""
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._select_collection.set_value(list(self._collections_names_map.keys()))
        if self._select_all_collections_checkbox is not None:
            self._select_all_collections_checkbox.check()
            self._select_collection.hide()

    def deselect_all(self) -> None:
        """Deselect all collections."""
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._select_collection.set_value([])
        if self._select_all_collections_checkbox is not None:
            self._select_all_collections_checkbox.uncheck()
            self._select_collection.show()



import os
from collections import defaultdict, namedtuple
from pathlib import Path
from typing import Generator, List, Union

import nrrd
import numpy as np

from supervisely import Api
from supervisely.collection.str_enum import StrEnum
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import ensure_base_path, get_file_ext, get_file_name
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.volume.volume import convert_3d_nifti_to_nrrd

VOLUME_NAME = "anatomic"
SCORE_NAME = "score"
LABEL_NAME = ["inference", "label", "annotation", "mask", "segmentation"]
MASK_PIXEL_VALUE = "Mask pixel value: "


class PlanePrefix(str, StrEnum):
    """Prefix for plane names."""

    CORONAL = "cor"
    SAGITTAL = "sag"
    AXIAL = "axl"


def read_cls_color_map(path: str) -> dict:
    """Read class color map from TXT file.

    ```txt
    1 Femur 255 0 0
    2 Femoral cartilage 0 255 0
    3 Tibia 0 0 255
    4 Tibia cartilage 255 255 0
    5 Patella 0 255 255
    6 Patellar cartilage 255 0 255
    7 Miniscus 175 175 175
    ```
    """

    cls_color_map = {}
    if not os.path.exists(path):
        return None
    try:
        with open(path, "r") as file:
            for line in file:
                parts = line.strip().split()
                color = None
                try:
                    color = list(map(int, parts[-3:]))
                except:
                    pass
                cls_id = int(parts[0])
                if not color:
                    cls_name = " ".join(parts[1:])
                else:
                    cls_name = " ".join(parts[1:-3])
                if cls_id in cls_color_map:
                    logger.warning(f"Duplicate class ID {cls_id} found in color map.")
                if cls_name in cls_color_map:
                    logger.warning(f"Duplicate class name {cls_name} found in color map.")
                if len(color) != 3:
                    logger.warning(f"Invalid color format for class {cls_name}. Expected 3 values.")
                if any(c < 0 or c > 255 for c in color):
                    logger.warning(
                        f"Invalid color value for class {cls_name}. Expected values between 0 and 255."
                    )
                cls_color_map[cls_id] = (cls_name, color)
    except Exception as e:
        logger.warning(f"Failed to read class color map from {path}: {e}")
        return None
    return cls_color_map


def read_json_map(path: str) -> dict:
    import json

    """Read JSON map from file."""
    if not os.path.exists(path):
        return None
    try:
        with open(path, "r") as file:
            json_map = json.load(file)
    except Exception as e:
        logger.warning(f"Failed to read JSON map from {path}: {e}")
        return None
    return json_map


def nifti_to_nrrd(nii_file_path: str, converted_dir: str) -> str:
    """Convert NIfTI 3D volume file to NRRD 3D volume file."""

    output_name = get_file_name(nii_file_path)
    if get_file_ext(output_name) == ".nii":
        output_name = get_file_name(output_name)

    data, header = convert_3d_nifti_to_nrrd(nii_file_path)

    nrrd_file_path = os.path.join(converted_dir, f"{output_name}.nrrd")
    ensure_base_path(nrrd_file_path)

    nrrd.write(nrrd_file_path, data, header)
    return nrrd_file_path


def get_annotation_from_nii(path: str) -> Generator[Mask3D, None, None]:
    """Get annotation from NIfTI 3D volume file."""

    data, header = convert_3d_nifti_to_nrrd(path)
    unique_classes = np.unique(data)

    for class_id in unique_classes:
        if class_id == 0:
            continue
        mask = Mask3D(data == class_id, volume_header=header)
        yield mask, class_id


def get_scores_from_table(csv_file_path: str, plane: str) -> dict:
    """Get scores from CSV table and return nested dictionary structure.

    Args:
        csv_file_path: Path to the CSV file containing layer scores

    Returns:
        Nested dictionary with structure:
        {
            "label_index": {
                "slice_index": {
                    "127": {
                        "score": float_value,
                        "comment": ""
                    }
                }
            }
        }
    """
    import csv

    if plane == PlanePrefix.CORONAL:
        plane = "0-1-0"
    elif plane == PlanePrefix.SAGITTAL:
        plane = "1-0-0"
    elif plane == PlanePrefix.AXIAL:
        plane = "0-0-1"

    result = defaultdict(lambda: defaultdict(dict))

    if not os.path.exists(csv_file_path):
        logger.warning(f"CSV file not found: {csv_file_path}")
        return result

    try:
        with open(csv_file_path, "r") as file:
            reader = csv.DictReader(file)
            label_columns = [col for col in reader.fieldnames if col.startswith("Label-")]

            for row in reader:
                frame_idx = int(row["Layer"]) - 1  # Assuming Layer is 1-indexed in CSV

                for label_col in label_columns:
                    label_index = int(label_col.split("-")[1])
                    score = f"{float(row[label_col]):.2f}"
                    result[label_index][plane][frame_idx] = {"score": score, "comment": None}

    except Exception as e:
        logger.warning(f"Failed to read CSV file {csv_file_path}: {e}")
        return {}

    return result


def _find_pixel_values(descr: str) -> int:
    """
    Find the pixel value in the description string.
    """
    lines = descr.split("\n")
    for line in lines:
        if line.strip().startswith(MASK_PIXEL_VALUE):
            try:
                value_part = line.strip().split(MASK_PIXEL_VALUE)[1]
                return int(value_part.strip())
            except (IndexError, ValueError):
                continue
    return None


def get_class_id_to_pixel_value_map(meta: ProjectMeta) -> dict:
    class_id_to_pixel_value = {}
    for obj_class in meta.obj_classes.items():
        pixel_value = _find_pixel_values(obj_class.description)
        if pixel_value is not None:
            class_id_to_pixel_value[obj_class.sly_id] = pixel_value
        elif "Segment_" in obj_class.name:
            try:
                pixel_value = int(obj_class.name.split("_")[-1])
                class_id_to_pixel_value[obj_class.sly_id] = pixel_value
            except (ValueError, IndexError):
                logger.warning(
                    f"Failed to parse pixel value from class name: {obj_class.name}. "
                    "Please ensure the class name ends with a valid integer."
                )
        else:
            logger.warning(
                f"Class {obj_class.name} does not have a pixel value defined in its description. "
                "Please update the class description to include 'Mask pixel value: <value>'."
            )
    return class_id_to_pixel_value


class AnnotationMatcher:
    def __init__(self, items, dataset_id):
        self._ann_paths = defaultdict(list)
        self._item_by_filename = {}
        self._item_by_path = {}

        self.items = items
        self._ds_id = dataset_id

        self._project_wide = False
        self._volumes = None

    @property
    def items(self):
        return self._items

    @items.setter
    def items(self, items):
        self._items = items
        self._ann_paths.clear()
        self._item_by_filename.clear()
        self._item_by_path.clear()

        for item in items:
            path = Path(item.ann_data)
            dataset_name = path.parts[-2]
            filename = path.name

            self._ann_paths[dataset_name].append(filename)
            self._item_by_filename[filename] = item
            self._item_by_path[(dataset_name, filename)] = item

    def get_volumes(self, api: Api):
        dataset_info = api.dataset.get_info_by_id(self._ds_id)
        if dataset_info.items_count > 0:
            datasets = {dataset_info.name: dataset_info}
            self._project_wide = False
        else:
            datasets = {
                dsinfo.name: dsinfo
                for dsinfo in api.dataset.get_list(dataset_info.project_id, recursive=True)
            }
            self._project_wide = True

        volumes = defaultdict(lambda: {})
        ds_filter = lambda ds_name: ds_name in self._ann_paths if self._project_wide else True
        for ds_name, ds_info in datasets.items():
            if ds_filter(ds_name):
                volumes[ds_name].update(
                    {info.name: info for info in api.volume.get_list(ds_info.id)}
                )

        if len(volumes) == 0:
            err_msg = "Failed to retrieve volumes from the project. Perhaps the input data structure is incorrect."
            raise RuntimeError(err_msg)

        self._volumes = volumes

    def match_items(self):
        """Match annotation files with corresponding volumes using regex-based matching."""
        item_to_volume = {}

        for dataset_name, volumes in self._volumes.items():
            volume_names = [parse_name_parts(name) for name in list(volumes.keys())]
            _volume_names = [vol for vol in volume_names if vol is not None]
            if len(_volume_names) == 0:
                logger.warning(f"No valid volume names found in dataset {dataset_name}.")
                continue
            elif len(_volume_names) != len(volume_names):
                logger.debug(f"Some volume names in dataset {dataset_name} could not be parsed.")
            volume_names = _volume_names

            ann_files = (
                self._ann_paths.get(dataset_name, [])
                if self._project_wide
                else list(self._ann_paths.values())[0]
            )
            for ann_file in ann_files:
                ann_name = parse_name_parts(ann_file)
                if ann_name is None:
                    logger.warning(f"Failed to parse annotation name: {ann_file}")
                    continue
                match = find_best_name_match(ann_name, volume_names)
                if match is not None:
                    if match.plane != ann_name.plane:
                        logger.warning(
                            f"Plane mismatch: {match.plane} != {ann_name.plane} for {ann_file}. Skipping."
                        )
                        continue
                    item_to_volume[self._item_by_filename[ann_file]] = volumes[match.full_name]

        volume_to_items = defaultdict(list)
        for item, volume in item_to_volume.items():
            volume_to_items[volume.id].append(item)

        for volume_id, items in volume_to_items.items():
            if len(items) == 1:
                items[0].is_semantic = True

        # items_to_remove = []
        for item, volume in item_to_volume.items():
            volume_shape = tuple(volume.file_meta["sizes"])
            if item.shape is None:
                continue
            if item.shape != volume_shape:
                logger.warning(f"Volume shape mismatch: {item.shape} != {volume_shape}")
                # items_to_remove.append(item)
        # for item in items_to_remove:
        # del item_to_volume[item]

        return item_to_volume

    def match_from_json(self, api: Api, json_map: dict):
        """
        Match annotation files with corresponding volumes based on a JSON map.

        Example json structure:
        {
            "cor_inference_1.nii": 123,
            "sag_mask_2.nii": 456
        }
        Where key is the annotation file name and value is the volume ID.

        For project-wide matching, the key should include dataset name:
        {
            "dataset1/cor_inference_1.nii": 123,
            "dataset2/sag_mask_2.nii": 456
        }
        """
        item_to_volume = {}

        for ann_path, volume_id in json_map.items():
            # Check if it's a project-wide path (contains dataset name)
            path_parts = Path(ann_path)
            if len(path_parts.parts) > 1:
                # Project-wide format: "dataset_name/filename.nii"
                dataset_name = path_parts.parts[-2]
                ann_name = path_parts.name
                item = self._item_by_path.get((dataset_name, ann_name))
            else:
                # Single dataset format: "filename.nii"
                ann_name = path_parts.name
                item = self._item_by_filename.get(ann_name)

            if item:
                volume = api.volume.get_info_by_id(volume_id)
                if volume:
                    item_to_volume[item] = volume

                    # Validate shape
                    volume_shape = tuple(volume.file_meta["sizes"])
                    if item.shape != volume_shape:
                        logger.warning(
                            f"Volume shape mismatch: {item.shape} != {volume_shape} for {ann_path}. Using anyway."
                        )
                else:
                    logger.warning(f"Volume {volume_id} not found for {ann_path}.")
            else:
                logger.warning(f"Item not found for annotation file {ann_path}.")

        # Set semantic flag for volumes with only one associated item
        volume_to_items = defaultdict(list)
        for item, volume in item_to_volume.items():
            volume_to_items[volume.id].append(item)
        for volume_id, items in volume_to_items.items():
            if len(items) == 1:
                items[0].is_semantic = True

        return item_to_volume


NameParts = namedtuple(
    "NameParts",
    [
        "full_name",
        "name_no_ext",
        "type",
        "plane",
        "is_ann",
        "patient_uuid",
        "case_uuid",
        "ending_idx",
    ],
)


def parse_name_parts(full_name: str) -> NameParts:
    from uuid import UUID

    name = get_file_name(full_name)
    if name.endswith(".nii"):
        name = get_file_name(name)
    name_no_ext = name

    type = None
    is_ann = False
    if VOLUME_NAME in full_name:
        type = VOLUME_NAME
    elif SCORE_NAME in full_name and full_name.endswith(".csv"):
        type = SCORE_NAME
    else:
        type = next((part for part in LABEL_NAME if part in full_name), None)
        is_ann = type is not None

    if type is None:
        return

    plane = None
    tokens = name_no_ext.lower().split("_")
    for part in PlanePrefix.values():
        if part in tokens:
            plane = part
            break

    if plane is None:
        return

    patient_uuid = None
    case_uuid = None

    if len(name_no_ext) > 73:
        try:
            uuids = name_no_ext[:73].split("_")
            if len(uuids) != 2:
                raise ValueError("Invalid UUID format")
            patient_uuid = UUID(name_no_ext[:36])
            case_uuid = UUID(name_no_ext[37:73])
        except ValueError:
            logger.debug(
                f"Failed to parse UUIDs from name: {name_no_ext}.",
                extra={"full_name": full_name},
            )
            patient_uuid = None
            case_uuid = None

    try:
        ending_idx = name_no_ext.split("_")[-1]
        if ending_idx.isdigit():
            ending_idx = int(ending_idx)
        else:
            ending_idx = None
    except ValueError:
        ending_idx = None
        logger.debug(
            f"Failed to parse ending index from name: {name_no_ext}.",
            extra={"full_name": full_name},
        )

    return NameParts(
        full_name=full_name,
        name_no_ext=name_no_ext,
        type=type,
        plane=plane,
        is_ann=is_ann,
        patient_uuid=patient_uuid,
        case_uuid=case_uuid,
        ending_idx=ending_idx,
    )


def find_best_name_match(item: NameParts, pool: List[NameParts]) -> Union[NameParts, None]:
    """
    Finds the best matching NameParts object from `pool` for the given annotation NameParts `item`.
    Prefers an exact match where all fields except `type` are the same, and `type` is 'anatomic'.
    Returns the matched NameParts object or None if not found.
    """
    pool_item_names = [i.full_name for i in pool]
    item_name = item.full_name
    # Prefer exact match except for type
    for i in pool:
        if i.name_no_ext == item.name_no_ext.replace(item.type, i.type):
            logger.debug(
                "Found exact match.",
                extra={"item": item_name, "pool_item": i.full_name},
            )
            return i

    logger.debug(
        "Failed to find exact match, trying to find a fallback match UUIDs.",
        extra={"item": item_name, "pool_items": pool_item_names},
    )

    # Fallback: match by plane and patient_uuid, type='anatomic'
    for i in pool:
        if (
            i.plane == item.plane
            and i.patient_uuid == item.patient_uuid
            and i.case_uuid == item.case_uuid
        ):
            logger.debug(
                "Found fallback match for item by UUIDs.",
                extra={"item": item_name, "i": i.full_name},
            )
            return i

    logger.debug(
        "Failed to find fallback match, trying to find a fallback match by plane.",
        extra={"item": item_name, "pool_items": pool_item_names},
    )

    # Fallback: match by plane and type='anatomic'
    for i in pool:
        if i.plane == item.plane:
            logger.debug(
                "Found fallback match for item by plane.",
                extra={"item": item_name, "i": i.full_name},
            )
            return i

    logger.debug(
        "Failed to find any match for item.",
        extra={"item": item_name, "pool_items": pool_item_names},
    )

    return None



def ply2pcd(input_path: str, output_path: str) -> None:
    """Convert .ply format to .pcd."""
    import open3d as o3d # pylint: disable=import-error

    points = o3d.io.read_point_cloud(input_path)
    o3d.io.write_point_cloud(output_path, points, write_ascii=True)


from typing import Dict, Any
from supervisely.app.widgets import Button, Card, ClassesTable, Container, Text


class ClassesSelector:
    title = "Classes Selector"
    description = "Select classes that will be used for inference"
    lock_message = "Select previous step to unlock"

    def __init__(self):
        # Init Step
        self.display_widgets = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Init Step Widgets
        self.classes_table = None
        # -------------------------------- #

        # Classes
        self.classes_table = ClassesTable()
        self.classes_table.hide()
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.classes_table])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        self.display_widgets.extend([self.validator_text, self.button])
        # -------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        self.card.lock()
        # -------------------------------- #

    @property
    def widgets_to_disable(self) -> list:
        return [self.classes_table]

    def load_from_json(self, data: Dict[str, Any]) -> None:
        if "classes" in data:
            self.set_classes(data["classes"])

    def get_selected_classes(self) -> list:
        return self.classes_table.get_selected_classes()

    def set_classes(self, classes) -> None:
        self.classes_table.select_classes(classes)

    def select_all_classes(self) -> None:
        self.classes_table.select_all()

    def get_settings(self) -> Dict[str, Any]:
        return {"classes": self.get_selected_classes()}

    def validate_step(self) -> bool:
        if self.classes_table.is_hidden():
            return True

        self.validator_text.hide()
        selected_classes = self.classes_table.get_selected_classes()
        n_classes = len(selected_classes)

        if n_classes == 0:
            self.validator_text.set(text="Please select at least one class", status="error")
            self.validator_text.show()
            return False

        class_word = "class" if n_classes == 1 else "classes"
        message_parts = [f"Selected {n_classes} {class_word}"]
        status = "success"
        is_valid = True

        self.validator_text.set(text=". ".join(message_parts), status=status)
        self.validator_text.show()
        return is_valid



from __future__ import annotations

import threading
from logging import Logger
from queue import Empty, Queue
from types import TracebackType
from typing import Callable, Optional, Type

from supervisely.sly_logger import logger


class Uploader:
    def __init__(
        self,
        upload_f: Callable,
        notify_f: Callable = None,
        exception_handler: Callable = None,
        logger: Logger = None,
    ):
        self._upload_f = upload_f
        self._notify_f = notify_f
        self._exception_handler = exception_handler
        if self._exception_handler is None:
            self._exception_handler = self._default_exception_handler
        self._logger = logger
        self.exception = None
        self._lock = threading.Lock()
        self._q = Queue()
        self._stop_event = threading.Event()
        self._exception_event = threading.Event()
        self._upload_thread = threading.Thread(
            target=self._upload_loop,
            daemon=True,
        )
        self.raise_from_notify = False
        self._notify_thread = None
        self._notify_q = Queue()
        if self._notify_f is not None:
            self._notify_thread = threading.Thread(target=self._notify_loop, daemon=True)
            self._notify_thread.start()
        self._upload_thread.start()

    def _notify_loop(self):
        try:
            while (
                not self._stop_event.is_set() or not self._notify_q.empty() or not self._q.empty()
            ):
                if self._exception_event.is_set():
                    return
                items = []
                try:
                    timeout = 0.1 if self._stop_event.is_set() else 1.0
                    item = self._notify_q.get(timeout=timeout)
                    items.append(item)
                    while True:
                        try:
                            items.append(self._notify_q.get_nowait())
                        except Empty:
                            break
                    if items:
                        self._notify_f(items)

                    for _ in range(len(items)):
                        self._notify_q.task_done()
                except Empty:
                    continue
        except StopIteration:
            self.stop()
            return
        except Exception as e:
            try:
                raise RuntimeError("Error in notify loop") from e
            except RuntimeError as e_:
                e = e_
            if self._logger is not None:
                self._logger.error("Error in notify loop: %s", str(e), exc_info=True)
            if self.raise_from_notify and not self._exception_event.is_set():
                self.set_exception(e)
            return

    def _upload_loop(self):
        try:
            while not self._stop_event.is_set() or not self._q.empty():
                if self._exception_event.is_set():
                    return
                items = []
                try:
                    timeout = 0.1 if self._stop_event.is_set() else 1.0
                    item = self._q.get(timeout=timeout)
                    items.append(item)
                    while True:
                        try:
                            items.append(self._q.get_nowait())
                        except Empty:
                            break
                    if items:
                        self._upload_f(items)
                        self.notify(items)

                    for _ in range(len(items)):
                        self._q.task_done()
                except Empty:
                    continue
        except StopIteration:
            self.stop()
            return
        except Exception as e:
            if self._logger is not None:
                self._logger.error("Error in upload loop: %s", str(e), exc_info=True)
            if not self._exception_event.is_set():
                self.set_exception(e)
            return

    def put(self, items):
        for item in items:
            self._q.put(item)

    def notify(self, items):
        if self._notify_thread is not None:
            for item in items:
                self._notify_q.put(item)

    def stop(self):
        self._stop_event.set()

    def join(self, timeout=None):
        self._upload_thread.join(timeout=timeout)
        if self._notify_thread is not None:
            self._notify_thread.join(timeout=timeout)

    def has_exception(self):
        return self._exception_event.is_set()

    def set_exception(self, exception: Exception):
        with self._lock:
            self._exception_event.set()
            self.exception = exception

    def __enter__(self):
        return self

    def _default_exception_handler(
        self,
        exception: Exception,
    ):
        raise exception

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        try:
            self.join(timeout=30)
            if self._upload_thread.is_alive():
                raise TimeoutError("Uploader thread didn't finish in time")
        except TimeoutError:
            _logger = logger
            if self._logger is not None:
                _logger = self._logger
            _logger.warning("Uploader thread didn't finish in time")
        if exc_type is not None:
            exc = exc_val.with_traceback(exc_tb)
            return self._exception_handler(exc)
        if self.has_exception():
            exc = self.exception
            try:
                raise RuntimeError(f"Error in uploader loop: {str(exc)}") from exc
            except Exception as exc:
                return self._exception_handler(exc)
        return False



from typing import Any, Dict, List

import yaml

from supervisely.app.widgets import (
    Button,
    Card,
    Checkbox,
    Container,
    Editor,
    Field,
    Input,
    Select,
    Text,
)


class InferenceMode:
    FULL_IMAGE = "Full Image"
    SLIDING_WINDOW = "Sliding Window"


class AddPredictionsMode:
    MERGE_WITH_EXISTING_LABELS = "Merge with existing labels"
    REPLACE_EXISTING_LABELS = "Replace existing labels"
    REPLACE_EXISTING_LABELS_AND_SAVE_IMAGE_TAGS = "Replace existing labels and save image tags"


class SettingsSelector:
    title = "Settings Selector"
    description = "Select additional settings for model inference"
    lock_message = "Select previous step to unlock"

    def __init__(self):
        # Init Step
        self.display_widgets: List[Any] = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Init Step Widgets
        self.inference_mode_selector = None
        self.inference_mode_field = None
        self.model_prediction_suffix_input = None
        self.model_prediction_suffix_field = None
        # self.model_prediction_suffix_checkbox = None
        self.predictions_mode_selector = None
        self.predictions_mode_field = None
        self.inference_settings = None
        # -------------------------------- #

        # Inference Mode
        self.inference_modes = [InferenceMode.FULL_IMAGE, InferenceMode.SLIDING_WINDOW]
        self.inference_mode_selector = Select(
            items=[Select.Item(mode) for mode in self.inference_modes]
        )
        self.inference_mode_selector.set_value(self.inference_modes[0])
        self.inference_mode_field = Field(
            content=self.inference_mode_selector,
            title="Inference mode",
            description="Select how to process images: full images or using sliding window.",
        )
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.inference_mode_field])
        # ----------------------------------- #

        # Class / Tag Suffix
        self.model_prediction_suffix_input = Input(
            value="_model", minlength=1, placeholder="Enter suffix e.g: _model"
        )
        self.model_prediction_suffix_field = Field(
            content=self.model_prediction_suffix_input,
            title="Class and tag suffix",
            description=(
                "Suffix that will be added to conflicting class and tag names. "
                "E.g. your project has a class 'person' with shape 'bitmap' and model has class 'person' with shape 'rectangle', "
                "then suffix will be added to the model predictions to avoid conflicts. E.g. 'person_model'."
            ),
        )
        # self.model_prediction_suffix_checkbox = Checkbox("Always add suffix to model predictions")
        # Add widgets to display ------------ #
        self.display_widgets.extend(
            [self.model_prediction_suffix_field]  # , self.model_prediction_suffix_checkbox]
        )
        # ----------------------------------- #

        # Prediction Mode
        self.prediction_modes = [
            AddPredictionsMode.MERGE_WITH_EXISTING_LABELS,
            AddPredictionsMode.REPLACE_EXISTING_LABELS,
            # AddPredictionsMode.REPLACE_EXISTING_LABELS_AND_SAVE_IMAGE_TAGS, # @TODO: Implement later
        ]
        self.predictions_mode_selector = Select(
            items=[Select.Item(mode) for mode in self.prediction_modes]
        )
        self.predictions_mode_selector.set_value(self.prediction_modes[0])
        self.predictions_mode_field = Field(
            content=self.predictions_mode_selector,
            title="Add predictions mode",
            description="Select how to add predictions to the project: by merging with existing labels or by replacing them.",
        )
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.predictions_mode_field])
        # ----------------------------------- #

        # Inference Settings
        self.inference_settings = Editor("", language_mode="yaml", height_px=300)
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.inference_settings])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.validator_text, self.button])
        # ----------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        self.card.lock()
        # ----------------------------------- #

    @property
    def widgets_to_disable(self) -> list:
        return [
            self.inference_mode_selector,
            self.model_prediction_suffix_input,
            # self.model_prediction_suffix_checkbox,
            self.predictions_mode_selector,
            self.inference_settings,
        ]

    def set_inference_settings(self, settings: Dict[str, Any]):
        if isinstance(settings, str):
            self.inference_settings.set_text(settings)
        else:
            self.inference_settings.set_text(yaml.safe_dump(settings))

    def get_inference_settings(self) -> Dict:
        settings = yaml.safe_load(self.inference_settings.get_text())
        if settings:
            return settings
        return {}

    def get_settings(self) -> Dict[str, Any]:
        return {
            "inference_mode": self.inference_mode_selector.get_value(),
            "model_prediction_suffix": self.model_prediction_suffix_input.get_value(),
            "predictions_mode": self.predictions_mode_selector.get_value(),
            "inference_settings": self.get_inference_settings(),
        }

    def load_from_json(self, data):
        inference_mode = data.get("inference_mode", None)
        if inference_mode:
            self.inference_mode_selector.set_value(inference_mode)

        model_prediction_suffix = data.get("model_prediction_suffix", None)
        if model_prediction_suffix is not None:
            self.model_prediction_suffix_input.set_value(model_prediction_suffix)

        predictions_mode = data.get("predictions_mode", None)
        if predictions_mode:
            self.predictions_mode_selector.set_value(predictions_mode)

        inference_settings = data.get("inference_settings", None)
        if inference_settings is not None:
            self.set_inference_settings(inference_settings)

    def validate_step(self) -> bool:
        return True



# coding: utf-8

import os

# Do NOT use directly for video extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_VIDEO_EXTENSIONS = ['.avi', '.mkv', '.mp4']


class VideoExtensionError(Exception):
    pass


def is_valid_ext(ext: str) -> bool:
    '''
    The function is_valid_ext checks file extension for list of supported video extensions('.avi', '.mp4')
    :param ext: file extention
    :return: True if file extention in list of supported images extensions, False - in otherwise
    '''
    return ext.lower() in ALLOWED_VIDEO_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    '''
    The function has_valid_ext checks if a given file has a supported extension('.avi', '.mp4')
    :param path: the path to the input file
    :return: True if a given file has a supported extension, False - in otherwise
    '''
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(ext: str):
    '''
    The function validate_ext generate exception error if file extention is not in list of supported videos
    extensions('.avi', '.mp4')
    :param ext: file extention
    '''
    if not is_valid_ext(ext):
        raise VideoExtensionError('Unsupported video extension: {}. Only the following extensions are supported: {}.'
                                  .format(ext, ALLOWED_VIDEO_EXTENSIONS))

